# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from functools import partial
from warnings import warn

import numpy as np

from pyxrd.generic.io import storables, Storable, get_case_insensitive_glob
from pyxrd.generic.models import DataModel
from pyxrd.generic.models.mixins import CSVMixin
from pyxrd.calculations.data_objects import AtomTypeData, AtomData
from pyxrd.calculations.atoms import get_atomic_scattering_factor, get_structure_factor

from mvc import Observer
from mvc.models.properties.string_properties import StringProperty
from mvc.models.properties.signal_mixin import SignalMixin
from mvc.models.properties.float_properties import FloatProperty
from mvc.models.properties.bool_property import BoolProperty
from mvc.models.properties.read_only_mixin import ReadOnlyMixin
from mvc.models.properties.labeled_property import LabeledProperty
from mvc.models.properties.integer_properties import IntegerProperty

@storables.register()
class AtomType(CSVMixin, DataModel, Storable):
    """
    An AtomType model contains all the physical & chemical information for 
    one ion, e.g. Fe3+ & Fe2+ are two different AtomTypes.
    """

    # MODEL METADATA:
    class Meta(DataModel.Meta):
        store_id = "AtomType"

    #: The project this AtomType belongs to or None. Effectively an alias for `parent`.
    project = property(DataModel.parent.fget, DataModel.parent.fset)

    _data_object = None
    @property
    def data_object(self):
        """
        The data object that is used in the calculations framework 
        (see :mod:`pyxrd.generic.calculations.atoms`).
        Is an instance of :class:`~pyxrd.generic.calculations.data_objects.AtomTypeData`
        """

        self._data_object.par_c = self.par_c
        self._data_object.debye = self.debye
        self._data_object.charge = self.charge
        self._data_object.weight = self.weight

        return self._data_object

    #: Name of the AtomType (e.g. :math:`Fe^{2+}`)
    name = StringProperty(
        default="", text="Name",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: The atomic number, or an arbitrarily high number (+300) for compounds
    atom_nr = IntegerProperty(
        default=0, text="Atom Nr",
        visible=True, persistent=True,
        widget_type="entry"
    )

    def __get_par_a(self, index=0):
        return self._data_object.par_a[index]

    def __set_par_a(self, value, index=0):
        assert (index >= 0 and index < 5)
        self._data_object.par_a[index] = value

    def __get_par_b(self, index=0):
        return self._data_object.par_b[index]

    def __set_par_b(self, value, index=0):
        assert (index >= 0 and index < 5)
        self._data_object.par_b[index] = value

    #: Atomic scattering factor :math:`a_1`
    par_a1 = FloatProperty(
        default=0.0, text="a1",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,),
        fget=partial(__get_par_a, index=0), fset=partial(__set_par_a, index=0)
    )
    #: Atomic scattering factor :math:`a_2`
    par_a2 = FloatProperty(
        default=0.0, text="a2",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,),
        fget=partial(__get_par_a, index=1), fset=partial(__set_par_a, index=1)
    )
    #: Atomic scattering factor :math:`a_3`
    par_a3 = FloatProperty(
        default=0.0, text="a3",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,),
        fget=partial(__get_par_a, index=2), fset=partial(__set_par_a, index=2)
    )
    #: Atomic scattering factor :math:`a_4`
    par_a4 = FloatProperty(
        default=0.0, text="a4",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,),
        fget=partial(__get_par_a, index=3), fset=partial(__set_par_a, index=3)
    )
    #: Atomic scattering factor :math:`a_5`
    par_a5 = FloatProperty(
        default=0.0, text="a5",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,),
        fget=partial(__get_par_a, index=4), fset=partial(__set_par_a, index=4)
    )

    #: Atomic scattering factor :math:`b_1`
    par_b1 = FloatProperty(
        default=0.0, text="b1",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,),
        fget=partial(__get_par_b, index=0), fset=partial(__set_par_b, index=0)
    )
    #: Atomic scattering factor :math:`b_2`
    par_b2 = FloatProperty(
        default=0.0, text="b2",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,),
        fget=partial(__get_par_b, index=1), fset=partial(__set_par_b, index=1)
    )
    #: Atomic scattering factor :math:`b_3`
    par_b3 = FloatProperty(
        default=0.0, text="b3",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,),
        fget=partial(__get_par_b, index=2), fset=partial(__set_par_b, index=2)
    )
    #: Atomic scattering factor :math:`b_4`
    par_b4 = FloatProperty(
        default=0.0, text="b4",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,),
        fget=partial(__get_par_b, index=3), fset=partial(__set_par_b, index=3)
    )
    #: Atomic scattering factor :math:`b_5`
    par_b5 = FloatProperty(
        default=0.0, text="b5",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,),
        fget=partial(__get_par_b, index=4), fset=partial(__set_par_b, index=4)
    )

    #: Atomic scattering factor c
    par_c = FloatProperty(
        default=0.0, text="c",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,)
    )

    #: Debye-Waller scattering factor
    debye = FloatProperty(
        default=0.0, text="c",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,)
    )

    #: The charge of the ion (eg. 3.0 for :math:`Al^{3+}`)
    charge = FloatProperty(
        default=0.0, text="Charge",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,)
    )

    #: The atomic weight for this ion
    weight = FloatProperty(
        default=0.0, text="Weight",
        visible=True, persistent=True,
        signal_name="data_changed", widget_type="entry",
        mix_with=(SignalMixin,)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Constructor takes any of its properties as a keyword argument.
            Any other arguments or keywords are passed to the base class.
        """
        keys = [ "data_%s" % prop.label for prop in type(self).Meta.get_local_persistent_properties()]
        keys.extend([ prop.label for prop in type(self).Meta.get_local_persistent_properties()])
        my_kwargs = self.pop_kwargs(kwargs, *keys)
        super(AtomType, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        # Set up data object
        self._data_object = AtomTypeData(
            par_a=np.zeros(shape=(5,), dtype=float),
            par_b=np.zeros(shape=(5,), dtype=float),
            par_c=0.0,
            debye=0.0,
            charge=0.0,
            weight=0.0
        )

        # Set attributes:
        self.name = str(self.get_kwarg(kwargs, "", "name", "data_name"))
        self.atom_nr = int(self.get_kwarg(kwargs, 0, "atom_nr", "data_atom_nr"))
        self.weight = float(self.get_kwarg(kwargs, 0, "weight", "data_weight"))
        self.charge = float(self.get_kwarg(kwargs, 0, "charge", "data_charge"))
        self.debye = float(self.get_kwarg(kwargs, 0, "debye", "data_debye"))

        for kw in ["par_a%d" % i for i in range(1, 6)] + ["par_b%d" % i for i in range(1, 6)] + ["par_c"]:
            setattr(
                self, kw, self.get_kwarg(kwargs, 0.0, kw, "data_%s" % kw)
            )

    def __str__(self):
        return "<AtomType %s (%s)>" % (self.name, id(self))

    def get_atomic_scattering_factors(self, stl_range):
        """
        Returns the atomic scattering factor for this `AtomType` for the given range
        of sin(theta)/lambda (`stl_range`) values. 
        """
        angstrom_range = ((stl_range * 0.05) ** 2)
        return get_atomic_scattering_factor(angstrom_range, self.data_object)

    pass # end of class

@storables.register()
class Atom(DataModel, Storable):
    """
    Atom objects combine structural information (z coordinate and proportion)
    and an AtomType. 
    """

    # MODEL METADATA:
    class Meta(DataModel.Meta):
        store_id = "Atom"
        layer_filters = [
            ("Layer file", get_case_insensitive_glob("*.lyr")),
        ]

    _data_object = None
    @property
    def data_object(self):
        """
        The data object that is used in the calculations framework 
        (see :mod:`pyxrd.generic.calculations.atoms`).
        Is an instance of :class:`~pyxrd.generic.calculations.data_objects.AtomData`  
        """
        self._data_object.default_z = self.default_z
        self._data_object.stretch_z = self.stretch_z
        self._data_object.pn = self.pn
        self._data_object.atom_type = getattr(self.atom_type, "data_object", None)
        return self._data_object

    component = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:

    #: The name of the Atom
    name = StringProperty(
        default="", text="Name",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: Default z coordinate for this Atom. Also see :attr:`~z`
    default_z = FloatProperty(
        default=0.0, text="Default z",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating whether or not z coordinates should be stretched
    #: using the silicate lattice and unit cell dimensions from the Component.
    #: Should be set for interlayer atoms, so their z coordinates are adjusted
    #: when the component basal spacing is changed. Also see `z`.
    stretch_z = BoolProperty(
        default=False, text="Stretch z values",
        visible=False, persistent=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    #: The z coordinate for this atom. If `stretch_values` is False or if
    #: this Atom's component is None, then this will return the `default_z`
    #: value. If `stretch_values` is True and a component is set on this Atom,
    #: it is calculated as::
    #:
    #:     `lattice_d + (default_z - lattice_d) * factor`
    #:
    #: where `lattice_d` and `factor` are given by calling
    #:`get_interlayer_stretch_factors` on the :class:`~pyxrd.phases.models.Component`.
    @FloatProperty(
        default=None, text="Z",
        visible=True, tabular=True,
        mix_with=(ReadOnlyMixin,)
    )
    def z(self):
        if self.stretch_values and self.component is not None:
            lattice_d, factor = self.component.get_interlayer_stretch_factors()
            return float(lattice_d + (self.default_z - lattice_d) * factor)
        return self.default_z

    #: The # of atoms (projected onto the c-axis for the considered unit cell)
    pn = FloatProperty(
        default=None, text="Multiplicity",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    def _get_weight(self):
        if self.atom_type is not None:
            return self.pn * self.atom_type.weight
        else:
            return 0.0

    #: The total weight for this Atom, taking `pn` into consideration.
    weight = FloatProperty(
        default=None, text="Weight", fget=_get_weight,
        visible=False, persistent=False,
        mix_with=(ReadOnlyMixin,)
    )

    _atom_type_uuid = None
    _atom_type_name = None
    def _set_atom_type(self, value):
        old = type(self).atom_type._get(self)
        if old is not None:
            self.relieve_model(old)
        type(self).atom_type._set(self, value)
        if value is not None:
            self.observe_model(value)

    #: The AtomType to be used for this Atom.
    atom_type = LabeledProperty(
        default=None, text="Atom Type",
        visible=True, persistent=False, tabular=True, data_type=AtomType,
        signal_name="data_changed",
        fset=_set_atom_type,
        mix_with=(SignalMixin,)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Constructor takes any of its properties as a keyword argument.
            
            In addition to the above, the constructor still supports the 
            following deprecated keywords, maping to a current keyword:
                - z: maps to the 'default_z' keyword.
                
            Any other arguments or keywords are passed to the base class.
        """

        my_kwargs = self.pop_kwargs(kwargs,
            "data_name", "data_z", "z", "data_pn", "data_atom_type", "stretch_z",
            "atom_type_uuid", "atom_type_name", "atom_type_index", "atom_type",
            *[prop.label for prop in type(self).Meta.get_local_persistent_properties()]
        )
        super(Atom, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        # Set up data object
        self._data_object = AtomData(
            default_z=0.0,
            pn=0.0
        )

        # Set attributes
        self.name = str(self.get_kwarg(kwargs, "", "name", "data_name"))

        self.stretch_values = bool(self.get_kwarg(kwargs, False, "stretch_values"))
        self.default_z = float(self.get_kwarg(kwargs, 0.0, "default_z", "data_z", "z"))
        self.pn = float(self.get_kwarg(kwargs, 0.0, "pn", "data_pn"))

        self.atom_type = self.get_kwarg(kwargs, None, "atom_type")
        self._atom_type_uuid = self.get_kwarg(kwargs, None, "atom_type_uuid")
        self._atom_type_name = self.get_kwarg(kwargs, None, "atom_type_name")

    def __str__(self):
        return "<Atom %s(%s)>" % (self.name, repr(self))

    def _unattach_parent(self):
        self.atom_type = None
        super(Atom, self)._unattach_parent()

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Observer.observe("removed", signal=True)
    def on_removed(self, model, prop_name, info):
        """
            This method observes the Atom types 'removed' signal,
            as such, if the AtomType gets removed from the parent project,
            it is also cleared from this Atom
        """
        if model == self.atom_type:
            self.atom_type = None

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_structure_factors(self, stl_range):
        """
        Get the atom's structure factor for a given range of 2*sin(θ) / λ values.
        Expects λ to be in nanometers!
        """
        if self.atom_type is not None:
            return float(get_structure_factor(stl_range, self.data_object))
        else:
            return 0.0

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def resolve_json_references(self):
        if getattr(self, "_atom_type_uuid", None) is not None:
            self.atom_type = type(type(self)).object_pool.get_object(self._atom_type_uuid)
        if getattr(self, "_atom_type_name", None) is not None:
            assert(self.component is not None)
            assert(self.component.phase is not None)
            assert(self.component.phase.project is not None)
            for atom_type in self.component.phase.project.atom_types:
                if atom_type.name == self._atom_type_name:
                    self.atom_type = atom_type
                    break
        self._atom_type_uuid = None
        self._atom_type_name = None

    def json_properties(self):
        retval = super(Atom, self).json_properties()
        if self.component is None or self.component.export_atom_types:
            retval["atom_type_name"] = self.atom_type.name if self.atom_type is not None else ""
        else:
            retval["atom_type_uuid"] = self.atom_type.uuid if self.atom_type is not None else ""
        return retval

    @staticmethod
    def get_from_csv(filename, callback=None, parent=None):
        """
        Returns a list of atoms fetched from the .CSV file `filename`.
        If parent is passed, this will be used to resolve AtomType references,
        and will be passed to the constructor of the Atom as a keyword.
        If callback is passes it will be called with the loaded atom as the
        first and only argument.
        """
        import csv
        atl_reader = csv.reader(open(filename, 'rb'), delimiter=',', quotechar='"') # TODO create csv dialect!
        header = True
        atoms = []

        types = dict()
        if parent is not None:
            for atom_type in parent.phase.project.atom_types:
                if not atom_type.name in types:
                    types[atom_type.name] = atom_type

        for row in atl_reader:
            if not header and len(row) >= 4:
                if len(row) == 5:
                    name, z, def_z, pn, atom_type = row[0], float(row[1]), float(row[2]), float(row[3]), types[row[4]] if parent is not None else None
                else:
                    name, z, pn, atom_type = row[0], float(row[1]), float(row[2]), types[row[3]] if parent is not None else None
                    def_z = z

                if atom_type in types:
                    atom_type = types[atom_type]

                new_atom = Atom(name=name, z=z, default_z=def_z, pn=pn, atom_type=atom_type, parent=parent)
                atoms.append(new_atom)
                if callback is not None and callable(callback):
                    callback(new_atom)
                del new_atom

            header = False
        return atoms

    @staticmethod
    def save_as_csv(filename, atoms):
        """
        Saves a list of atoms to the passed filename.
        """
        import csv
        atl_writer = csv.writer(open(filename, 'wb'), delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        atl_writer.writerow(["Atom", "z", "def_z", "pn", "Element"])
        for item in atoms:
            if item is not None and item.atom_type is not None:
                atl_writer.writerow([item.name, item.z, item.default_z, item.pn, item.atom_type.name])

    pass # end of class




# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk  # @UnresolvedImport


import os
import csv

################################################################################
# Array-like item repositioning:
################################################################################
def repos(ln, old_pos, new_pos):
    """
        Return a new list in which each item contains the index of the item
        in the old order.
        Uses the ranges approach (best for large arrays).
    """
    lb = min(new_pos, old_pos)
    ub = max(new_pos, old_pos)
    adj_range = []
    if new_pos < old_pos:
        adj_range.append(old_pos)
        adj_range += list(range(lb, ub))
    else:
        adj_range += list(range(lb + 1, ub + 1))
        adj_range.append(old_pos)
    return list(range(0, lb)) + adj_range + list(range(ub, ln - 1))

def simple_repos(ln, old_pos, new_pos):
    """
        Return a new list in which each item contains the index of the item
        in the old order.
        Uses the delete/insert approach (best for small arrays).
    """
    r1 = list(range(ln))
    val = r1[old_pos]
    del r1[old_pos]
    r1.insert(new_pos, val)
    return r1

def smart_repos(ln, old_pos, new_pos):
    """
        Return a new list in which each item contains the index of the item
        in the old order.
        Decides which algorithm to use based on the size of the arrays.
    """
    if ln > 65:
        return repos(ln, old_pos, new_pos)
    else:
        return simple_repos(ln, old_pos, new_pos)

################################################################################
# Treestore creation from filesystem
################################################################################
def create_valuestore_from_file(filename, data_type=float):
    liststore = Gtk.ListStore(str, data_type)
    with open(filename, 'r') as f:
        reader = csv.reader(f)
        next(reader) # skip header
        for row in reader:
            row[1] = data_type(row[1])
            liststore.append(row)
    return liststore

def create_treestore_from_directory(directory):
    treestore = Gtk.TreeStore(str, str, bool)
    treestore.append(None, ("", "", True))
    parents = {}
    for root, dirnames, filenames in os.walk(directory):
        for dirname in dirnames:
            parents[os.path.join(root, dirname)] = treestore.append(parents.get(root, None), (dirname, "", False))
        for filename in filenames:
            treestore.append(parents.get(root, None), (os.path.splitext(filename)[0], "%s/%s" % (root, filename), True))
    return treestore



#!/usr/bin/python
# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

"""
    Tools for making scripting easier
"""

def reload_settings(clear_script=True):
    """
        This will reload the PyXRD settings after clearing the script path from
        the command line arguments. This allows to run the GUI when needed.
    """
    import sys
    from copy import copy

    # Make a copy to prevent errors
    args = copy(sys.argv)
    for i, arg in enumerate(args):
        if arg == "-s":
            # Clear the next key (contains the script name
            del sys.argv[i + 1]
            # Clear the flag
            del sys.argv[i]
            # Exit the loop
            break

    from pyxrd.data import settings
    settings.SETTINGS_APPLIED = False # clear this flag to reload settings
    settings.initialize() # reload settings

def launch_gui(project=None):
    """ Launches the GUI, you should run reload_settings before calling this! """
    from pyxrd.core import _run_gui
    _run_gui(project=project) # launch gui


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#
# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import scipy.stats
from math import sqrt

from mvc.models.properties.float_properties import FloatProperty

from pyxrd.generic.io.custom_io import storables
from pyxrd.refinement.refinables.properties import RefinableMixin

from .insitu_behaviour import InSituBehaviour

BEHAVIOUR_CLASS = "SSSR0Behaviour"

@storables.register()
class SSSR0Behaviour(InSituBehaviour):
    """
    	Behaviour class for R0 Smectite with 3 components
    """

    # MODEL INTEL:
    class Meta(InSituBehaviour.Meta):
        store_id = "SSSR0Behaviour"
        concrete = True # indicates this can be instantiated and added in the UI

    cation_hydration_factor = FloatProperty(
        description="Related to the cation hydration enthalpy",
        default=0.54, text="Cation hydration factor",
        minimum=0.0, maximum=2.0,
        refinable=True, persistent=True, visible=True,
        mix_with=(RefinableMixin,)
    )
     
    min_swelling_layer_charge = FloatProperty(
        description="The minimum layer charge required for swelling",
        default=-0.1, text="Minimum swelling layer charge",
        minimum = -2, maximum=0,
        refinable=True, persistent=True, visible=True,
        mix_with=(RefinableMixin,)
    )
      
    layer_charge_mean = FloatProperty(
        description="The mean of the normal layer charge distribution",
        default=-0.4, text="Mean layer charge",
        minimum = -2, maximum=0,
        refinable=True, persistent=True, visible=True,
        mix_with=(RefinableMixin,)
    )

    layer_charge_stdev = FloatProperty(
        default=0.05, text="The standard deviation of the layer charges",
        minimum = 0.0001, maximum=0.75,
        refinable=True, persistent=True, visible=True,
        mix_with=(RefinableMixin,)
    )
        
    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        my_kwargs = self.pop_kwargs(kwargs,
            *[prop.label for prop in SSSR0Behaviour.Meta.get_local_persistent_properties()]
        )
        super(SSSR0Behaviour, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        with self.data_changed.hold():
            self.cation_hydration_factor = self.get_kwarg(kwargs, self.cation_hydration_factor, "cation_hydration_factor")
            self.min_swelling_layer_charge = self.get_kwarg(kwargs, self.min_swelling_layer_charge, "min_swelling_layer_charge")
            self.layer_charge_mean = self.get_kwarg(kwargs, self.layer_charge_mean, "layer_charge_mean")
            self.layer_charge_stdev = self.get_kwarg(kwargs, self.layer_charge_stdev, "layer_charge_stdev")
            
        pass #end of constructor

            
    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_layer_charge_distribution(self, mean = -0.4, stdev = 0.05):        
        distr = scipy.stats.norm(mean, stdev)
        return distr

    def get_layer_type_distribution(self, distr, boundaries):
        """
            Returns a 3-tuple with the % of layers having:
            0 layers of water
            1 layers of water
            2 layers of water
        """
        (B0to1, B1to2, B2to0) = boundaries
        cdfB0to1 = distr.cdf(B0to1) # Get # layers with B0to1 or smaller layer charge 
        cdfB1to2 = distr.cdf(B1to2) # Get # layers with B1to2 or smaller layer charge
        cdfB2to0 = distr.cdf(B2to0) # Get # layers with B2to0 or smaller layer charge
        total = distr.cdf(0)        # Get # layers with B2 or smaller layer charge
        # returns fraction of 0w, 1w and 2w layers:
        return (cdfB0to1+total-cdfB2to0)/total,(cdfB1to2-cdfB0to1)/total,(cdfB2to0-cdfB1to2)/total
        
    def get_layer_type_boundaries(self, RH):
        """
            Gets the layer type boundaries (expressed as layer charge)
            in function of RH (expressed as a fraction)
        """
        RH_step_size = (self.cation_hydration_factor*sqrt(RH))
        b2to0_RH_factor = self.min_swelling_layer_charge
        b1to2_RH_factor = b2to0_RH_factor - RH_step_size
        b0to1_RH_factor = b1to2_RH_factor - RH_step_size
        
        return b0to1_RH_factor, b1to2_RH_factor, b2to0_RH_factor
    
    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def apply(self, phase, RH):
        super(SSSR0Behaviour, self).apply(phase)

        RH = RH / 100.
        
        print("Applying SSSR0Behaviour to %s" % phase, "For RH: %.2f" % RH)

        # Get layer charge coundaries for the given relative humidity
        boundaries = self.get_layer_type_boundaries(RH)
        
        print(" boundaries:", boundaries)

        # Get the layer charge distribution for the given phase
        lc_distr = self.get_layer_charge_distribution(self.layer_charge_mean, self.layer_charge_stdev)
        print(" lc_distr:", lc_distr)
        
        # Calculate the layer type distribution for the calculated boundaries and distribution
        W0, W1, W2 = self.get_layer_type_distribution(lc_distr, boundaries)
        print(" W0:", W0)
        print(" W1:", W1)
        print(" W2:", W2)
        
        # Set probability model factors:
        phase.probabilities.F0 = W2
        phase.probabilities.F1 = W1/(W1+W0)
        
        # TODO update d-spacings
                
    def is_compatible_with(self, phase):
        # TODO check names of components
        try:
            return (phase.R == 0 and phase.G == 3)
        except:
            return False   
        
    pass



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

"""
    A python implementation of the matrix algorithm developed for the X-ray
        diffraction analysis of disordered lamellar structures
"""

import pkg_resources
from pyxrd.__version import __version__



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import json
import numpy as np

class PyXRDEncoder(json.JSONEncoder):
    """
        A custom JSON encoder that checks if:
            - the object has a to_json callable method, if so it is called to
              convert the object to a JSON encodable object.
              E.g. the default implementation from the Storable class is a dict object
              containing:
               - a 'type' key mapped to the storage type id of the storable class
               - a 'properties' key mapped to a dict of name-values for each
                 property that needs to be stored in order to be able to 
                 recreate the object.
              If the user registered this class as a storable
              (see the 'registes_storable' method or the Storable class)
              then the JSON object is transformed back into the actual Python
              object using its from_json(...) method. Default implementation
              finds the registered class using the 'type' value and passes the
              'properties' value to its constructor as keyword arguments.  
            - if the object is a numpy array, it is converted to a list
            - if the object is a wrapped list, dictionary, ... (ObsWrapper 
              subclass) then the wrapped object is returned, as these are
              directly JSON encodable.
            - fall back to the default JSONEncoder methods
    """

    ###########################################################################
    # Convenience functions: use these!
    ###########################################################################
    @classmethod
    def dump_object(cls, obj):
        """ Serialize an object using this encoder and return it as a string """
        return json.dumps(obj, indent=4, cls=cls)

    @classmethod
    def dump_object_to_file(cls, obj, f):
        """ Serialize an object using this encoder and dump it into a file"""
        return json.dump(obj, f, indent=4, cls=cls)

    ###########################################################################
    # Sub class implementation:
    ###########################################################################
    def default(self, obj):
        from mvc.support.observables import ObsWrapper
        if hasattr(obj, "to_json") and callable(getattr(obj, "to_json")):
            return obj.to_json()
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, ObsWrapper):
            return obj._obj # return the wrapped object
        return json.JSONEncoder(self).default(obj)


class PyXRDDecoder(json.JSONDecoder):
    """
        A custom JSON decoder that can decode objects, following these steps:
            - decode the JSON object at once using the default decoder
            - the resulting dict is then parsed:
               - if a valid 'type' and a 'properties' key is given,
                 the object is translated using the mapped class type's 
                 'from_json' method. This mapping is done using a dict mapping
                 the json class name to an actual class type (`mapper` __init__ keyword)
               - parent keyword arguments are passed on (e.g. a project) to
                 the from_json method as well
    """

    def __init__(self, mapper=None, parent=None, **kwargs):
        super(PyXRDDecoder, self).__init__(**kwargs)
        self.mapper = mapper
        self.parent = parent

    ###########################################################################
    # Convenience functions: use these!
    ###########################################################################
    @classmethod
    def decode_file(cls, f, mapper, parent=None):
        data = f.read()
        data = data.decode("utf-8") if isinstance(data, bytes) else data 
        return json.loads(data, cls=PyXRDDecoder, mapper=mapper, parent=parent)

    @classmethod
    def decode_string(cls, string, mapper, parent=None):
        return json.loads(string, cls=PyXRDDecoder, mapper=mapper, parent=parent)

    ###########################################################################
    # Sub class implementation:
    ###########################################################################
    def decode(self, string):
        """ Decodes a json string into an object """
        # First use a regular decode:
        obj = super(PyXRDDecoder, self).decode(string)
        # Then parse this dict into an actual python object:
        return self.__pyxrd_decode__(obj) or obj

    def __pyxrd_decode__(self, obj, **kwargs):
        """ Decodes the PyXRD JSON object serialization """
        if isinstance(obj, list):
            for index, subobj in enumerate(obj):
                obj[index] = self.__pyxrd_decode__(subobj) or subobj
            return obj
        if "type" in obj:
            objtype = self.mapper[obj["type"]]
            if "properties" in obj and hasattr(objtype, "from_json"):
                if self.parent is not None and not "parent" in kwargs:
                    kwargs["parent"] = self.parent
                return objtype.from_json(**dict(obj["properties"], **kwargs))
        logger.warn("__pyxrd_decode__ will return None for %s!" % str(obj)[:30] + "..." + str(obj)[:-30])
        return None



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

class TreeNode(object):
    """
        A TreeNode can be used to build a Tree of objects with parents & 
        children and support for inserting, appending, removing, iterating and
        inspecting them.
        This is the type that models should use.
        The root object can than be wrapped easily by a UI-toolkit specific 
        wrapper, like the gtk ObjectTreeStore wrapper. 
    """
    def insert(self, index, child_node):
        """
            Inserts the 'child_node' in this TreeNode at the given index.
            If the passed argument is not a TreeNode instance, it is wrapped
            in one. So you can safely pass in arbitrary objects.
            This functions returns the (wrapped) child_node.
        """
        assert isinstance(child_node, type(self))
        child_node.parent = self
        self._children.insert(index, child_node)
        self.on_grandchild_inserted(child_node)
        return child_node

    def append(self, child_node):
        """
            Appends the 'child_node' to this TreeNode. If the passed
            argument is not a TreeNode instance, it is wrapped in one.
            So you can safely pass in arbitrary objects.
            This functions returns the (wrapped) child_node.
        """
        assert isinstance(child_node, TreeNode)
        return self.insert(len(self._children), child_node)

    def remove(self, child_node):
        """
            Removes the given child_node from this TreeNode's list of children.
            'child_node' *must* be a TreeNode instance. 
        """
        assert isinstance(child_node, type(self))
        self.on_grandchild_removed(child_node)
        self._children.remove(child_node)
        child_node._parent = None

    def on_grandchild_removed(self, child_node):
        if self.parent is not None:
            self.parent.on_grandchild_removed(child_node)

    def on_grandchild_inserted(self, child_node):
        if self.parent is not None:
            self.parent.on_grandchild_removed(child_node)

    _parent = None
    @property
    def parent(self):
        return self._parent
    @parent.setter
    def parent(self, parent):
        if self._parent:
            self._parent.remove(self)
        self._parent = parent

    @property
    def has_children(self):
        return bool(self._children)

    @property
    def child_count(self):
        return len(self._children)

    def __init__(self, obj=None, children=()):
        super(TreeNode, self).__init__()
        self.object = obj
        self._children = list()
        for child in children: self.append(child)

    def get_child_node_index(self, child_node):
        return self._children.index(child_node)

    def get_root_node(self):
        parent = self.parent
        while parent.parent is not None:
            parent = parent.parent
        return parent

    def get_next_node(self):
        try:
            return self.parent.get_child_node(
                self.parent.get_child_node_index(self) + 1)
        except IndexError:
            return None

    def get_prev_node(self):
        try:
            return self.parent.get_child_node(
                self.parent.get_child_node_index(self) - 1)
        except IndexError:
            return None

    def get_indices(self):
        parent = self.parent
        node = self
        indeces = tuple()
        while parent is not None:
            indeces += (parent.get_child_node_index(node),)
            node = parent
            parent = parent.parent
        return indeces[::-1] or None

    def get_child_node(self, *indeces):
        node = self
        for index in indeces:
            try:
                node = node._children[index]
            except IndexError:
                return None
        return node

    def get_first_child_node(self):
        try:
            return self._children[0]
        except IndexError:
            return None

    def get_child_object(self, *indeces):
        return self.get_child_node(*indeces).object

    def clear(self):
        for c in self.iter_children(reverse=True):
            c.parent = None

    def iter_children(self, reverse=False, recursive=True):
        children = self._children if not reverse else self._children[::-1]
        for child_node in children:
            if recursive and child_node.has_children:
                for c in child_node.iter_children(reverse=reverse, recursive=recursive):
                    yield c
            yield child_node

    def __repr__(self):
        return '%s(%s - %s)' % (type(self).__name__, self.object, "%d child nodes" % len(self._children))



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np

from .R0models import R0G1Model, R0G2Model, R0G3Model, R0G4Model, R0G5Model, R0G6Model # @UnusedImport
from .R1models import R1G2Model, R1G3Model, R1G4Model # @UnusedImport
from .R2models import R2G2Model, R2G3Model # @UnusedImport
from .R3models import R3G2Model # @UnusedImport

# Overview of what is:
#   x   = currently implemented
#   np  = not possible
#   -/o = not yet implemented
#   o   = priority
#
#       G1  G2  G3  G4  G5  G6
#   R0  x   x   x   x   x   x
#   R1  np  x   x   x   -   -
#   R2  np  x   x   -   -   -
#   R3  np  x   -   -   -   -

RGbounds = np.array([
    [1, 1, 1, 1, 1, 1],
    [-1, 1, 1, 1, 0, 0],
    [-1, 1, 1, 0, 0, 0],
    [-1, 1, 0, 0, 0, 0],
])

def get_Gbounds_for_R(R, G):
    global RGbounds
    maxR, maxG = RGbounds.shape
    low, upp = 1, 6
    if R >= 0 and R < maxR:
        bounds = RGbounds[R]
        low, upp = 1 + np.argmax(bounds == 1), maxG - np.argmax(bounds[::-1] == 1)
    else:
        raise ValueError("Cannot yet handle R%d!" % R)
    return (low, upp, max(min(G, upp), low))

def get_Rbounds_for_G(G, R):
    global RGbounds
    maxR, maxG = RGbounds.shape
    low, upp = 0, 0
    if G >= 1 and G <= maxG:
        bounds = RGbounds[:, G - 1]
        low, upp = np.argmax(bounds == 1), maxR - np.argmax(bounds[::-1] == 1) - 1
    else:
        raise ValueError("Cannot yet handle %d layer structures!" % G)
    return (low, upp, max(min(R, upp), low))

def get_correct_probability_model(R, G):
    global RGbounds
    if (RGbounds[R, G - 1] > 0):
        return globals()["R%dG%dModel" % (R, G)]
    else:
        raise ValueError("Cannot (yet) handle R%d for %d layer structures!" % (R, G))



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .atom_relation_controllers import (
    EditUnitCellPropertyController,
    EditAtomRatioController,
    EditAtomContentsController,
    ContentsListController,
    EditAtomRelationsController
)

from .CSDS_controllers import (
    EditCSDSTypeController,
    EditCSDSDistributionController
)

from .layer_controllers import EditLayerController

from .component_controllers import (
    ComponentsController,
    EditComponentController
)

from .add_phase_controller import AddPhaseController
from .edit_phase_controller import EditPhaseController
from .raw_pattern_phase_controller import EditRawPatternPhaseController
from .phases_controller import PhasesController


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc import Controller

from pyxrd.generic.controllers import BaseController

from pyxrd.probabilities.views import get_correct_probability_views
from pyxrd.probabilities.models import RGbounds

def get_correct_probability_controllers(probability, parent_controller, independents_view, dependents_view):
    if probability is not None:
        G = probability.G
        R = probability.R
        if (RGbounds[R, G - 1] > 0):
            return BaseController(model=probability, parent=parent_controller, view=independents_view), \
                   MatrixController(current=R, model=probability, parent=parent_controller, view=dependents_view)
        else:
            raise ValueError("Cannot (yet) handle R%d for %d layer structures!" % (R, G))

class EditProbabilitiesController(BaseController):

    independents_view = None
    matrix_view = None
    auto_adapt = False

    def __init__(self, *args, **kwargs):
        BaseController.__init__(self, *args, **kwargs)
        self._init_views(kwargs["view"])
        self.update_views()

    def _init_views(self, view):
        self.independents_view, self.dependents_view = get_correct_probability_views(self.model, view)
        self.independents_controller, self.dependents_controller = get_correct_probability_controllers(self.model, self, self.independents_view, self.dependents_view)
        view.set_views(self.independents_view, self.dependents_view)

    @BaseController.model.setter
    def model(self, model):
        if self.view is not None:
            self.independents_controller.model = None # model
            self.dependents_controller.model = None # model
        super(EditProbabilitiesController, self)._set_model(model)
        if self.view is not None:
            self.independents_controller.model = model
            self.dependents_controller.model = model
            self.update_views()

    def update_views(self): # needs to be called whenever an independent value changes
        with self.model.data_changed.hold():
            self.dependents_view.update_matrices(self.model)
            self.independents_view.update_matrices(self.model)

    def register_adapters(self):
        return

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Controller.observe("data_changed", signal=True)
    def notif_updated(self, model, prop_name, info):
        self.update_views()
        return

    pass # end of class

class MatrixController(BaseController):
    auto_adapt = False

    def __init__(self, current, *args, **kwargs):
        BaseController.__init__(self, *args, **kwargs)
        self.current_W = current
        self.current_P = current

    def register_adapters(self):
        return

    def on_w_prev_clicked(self, widget, *args):
        self.current_W = self.view.show_w_matrix(self.current_W - 1)
    def on_w_next_clicked(self, widget, *args):
        self.current_W = self.view.show_w_matrix(self.current_W + 1)

    def on_p_prev_clicked(self, widget, *args):
        self.current_P = self.view.show_p_matrix(self.current_P - 1)
    def on_p_next_clicked(self, widget, *args):
        self.current_P = self.view.show_p_matrix(self.current_P + 1)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from warnings import warn
from random import choice

from mvc import Observer
from mvc.observers import ListObserver
from mvc.models.properties import (
    StringProperty, SignalMixin, BoolProperty, LabeledProperty,
    FloatProperty, ListProperty
)

from pyxrd.generic.io import storables, get_case_insensitive_glob

from pyxrd.generic.models.properties import InheritableMixin, ObserveChildMixin
from pyxrd.refinement.refinables.properties import RefinableMixin
from pyxrd.refinement.refinables.mixins import RefinementGroup
from pyxrd.refinement.refinables.metaclasses import PyXRDRefinableMeta

from pyxrd.probabilities.models import get_correct_probability_model

from .abstract_phase import AbstractPhase
from .CSDS import DritsCSDSDistribution
from .component import Component

@storables.register()
class Phase(RefinementGroup, AbstractPhase, metaclass=PyXRDRefinableMeta):

    # MODEL INTEL:
    class Meta(AbstractPhase.Meta):
        store_id = "Phase"
        file_filters = [
            ("Phase file", get_case_insensitive_glob("*.PHS")),
        ]

    _data_object = None
    @property
    def data_object(self):
        self._data_object.type = "Phase"
        self._data_object.valid_probs = (all(self.probabilities.P_valid) and all(self.probabilities.W_valid))

        if self._data_object.valid_probs:
            self._data_object.sigma_star = self.sigma_star
            self._data_object.CSDS = self.CSDS_distribution.data_object

            self._data_object.G = self.G
            self._data_object.W = self.probabilities.get_distribution_matrix()
            self._data_object.P = self.probabilities.get_probability_matrix()

            self._data_object.components = [None] * len(self.components)
            for i, comp in enumerate(self.components):
                self._data_object.components[i] = comp.data_object
        else:
            self._data_object.sigma_star = None
            self._data_object.CSDS = None
            self._data_object.G = None
            self._data_object.W = None
            self._data_object.P = None
            self._data_object.components = None

        return self._data_object

    project = property(AbstractPhase.parent.fget, AbstractPhase.parent.fset)

    # PROPERTIES:

    #: Flag indicating whether the CSDS distribution is inherited from the
    #: :attr:`based_on` phase or not.
    @BoolProperty(
        default=False, text="Inh. mean CSDS",
        visible=True, persistent=True, tabular=True
    )
    def inherit_CSDS_distribution(self):
        return self._CSDS_distribution.inherited
    @inherit_CSDS_distribution.setter
    def inherit_CSDS_distribution(self, value):
        self._CSDS_distribution.inherited = value

    #: Flag indicating whether to inherit the display color from the
    #: :attr:`based_on` phase or not.
    inherit_display_color = BoolProperty(
        default=False, text="Inh. display color",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating whether to inherit the sigma start value from the
    #: :attr:`based_on` phase or not.
    inherit_sigma_star = BoolProperty(
        default=False, text="Inh. sigma star",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    _based_on_index = None # temporary property
    _based_on_uuid = None # temporary property

    #: The :class:`~Phase` instance this phase is based on
    based_on = LabeledProperty(
        default=None, text="Based on phase",
        visible=True, persistent=False, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin, ObserveChildMixin)
    )
    @based_on.setter
    def based_on(self, value):
        old = type(self).based_on._get(self)
        if value == None or value.get_based_on_root() == self or value.parent != self.parent:
            value = None
        if value != old:
            type(self).based_on._set(self, value)
            for component in self.components:
                component.linked_with = None

    # INHERITABLE PROPERTIES:

    #: The sigma star orientation factor
    sigma_star = FloatProperty(
        default=3.0, text="σ* [°]", math_text="$\sigma^*$ [°]",
        minimum=0.0, maximum=90.0,
        visible=True, persistent=True, tabular=True, refinable=True,
        inheritable=True, inherit_flag="inherit_sigma_star", inherit_from="based_on.sigma_star",
        signal_name="data_changed",
        mix_with=(SignalMixin, RefinableMixin, InheritableMixin)
    )

    # A :class:`~pyxrd.phases.models.CSDS` instance
    CSDS_distribution = LabeledProperty(
        default=None, text="CSDS Distribution",
        visible=True, persistent=True, tabular=True, refinable=True,
        inheritable=True, inherit_flag="inherit_CSDS_distribution", inherit_from="based_on.CSDS_distribution",
        signal_name="data_changed",
        mix_with=(SignalMixin, RefinableMixin, InheritableMixin, ObserveChildMixin)
    )

    # A :class:`~pyxrd._probabilities.models._AbstractProbability` subclass instance
    probabilities = LabeledProperty(
        default=None, text="Probablities",
        visible=True, persistent=True, tabular=True, refinable=True,
        signal_name="data_changed",
        mix_with=(SignalMixin, RefinableMixin, ObserveChildMixin)
    )
    @probabilities.setter
    def probabilities(self, value):
        type(self).probabilities._set(self, value)
        if value is not None:
            value.update()

    #: The color this phase's X-ray diffraction pattern should have.
    display_color = StringProperty(
        fset=AbstractPhase.display_color.fset,
        fget=AbstractPhase.display_color.fget,
        fdel=AbstractPhase.display_color.fdel,
        doc=AbstractPhase.display_color.__doc__,
        default="#008600", text="Display color",
        visible=True, persistent=True, tabular=True, widget_type='color',
        inheritable=True, inherit_flag="inherit_display_color", inherit_from="based_on.display_color",
        signal_name="visuals_changed",
        mix_with=(SignalMixin, InheritableMixin)
    )

    #: The list of components this phase consists of
    components = ListProperty(
        default=None, text="Components",
        visible=True, persistent=True, tabular=True, refinable=True,
        widget_type="custom", data_type=Component,
        mix_with=(RefinableMixin,)
    )

    #: The # of components
    @AbstractPhase.G.getter
    def G(self):
        if self.components is not None:
            return len(self.components)
        else:
            return 0

    #: The # of components
    @AbstractPhase.R.getter
    def R(self):
        if self.probabilities:
            return self.probabilities.R

    # Flag indicating whether or not the links (based_on and linked_with) should
    # be saved as well.
    save_links = True

    # REFINEMENT GROUP IMPLEMENTATION:
    @property
    def refine_title(self):
        return self.name

    @property
    def refine_descriptor_data(self):
        return dict(
            phase_name=self.refine_title,
            component_name="*"
        )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):

        my_kwargs = self.pop_kwargs(kwargs,
            "data_CSDS_distribution", "data_sigma_star", "data_components",
            "data_G", "G", "data_R", "R",
            "data_probabilities", "based_on_uuid", "based_on_index",
            "inherit_probabilities",
            *[prop.label for prop in Phase.Meta.get_local_persistent_properties()]
        )
        super(Phase, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        with self.data_changed.hold():

            CSDS_distribution = self.get_kwarg(kwargs, None, "CSDS_distribution", "data_CSDS_distribution")
            self.CSDS_distribution = self.parse_init_arg(
                CSDS_distribution, DritsCSDSDistribution, child=True,
                default_is_class=True, parent=self
            )
            self.inherit_CSDS_distribution = self.get_kwarg(kwargs, False, "inherit_CSDS_distribution")

            self.display_color = self.get_kwarg(kwargs, choice(self.line_colors), "display_color")
            self.inherit_display_color = self.get_kwarg(kwargs, False, "inherit_display_color")

            self.sigma_star = self.get_kwarg(kwargs, self.sigma_star, "sigma_star", "data_sigma_star")
            self.inherit_sigma_star = self.get_kwarg(kwargs, False, "inherit_sigma_star")

            self.components = self.get_list(kwargs, [], "components", "data_components", parent=self)

            G = int(self.get_kwarg(kwargs, 1, "G", "data_G"))
            R = int(self.get_kwarg(kwargs, 0, "R", "data_R"))
            if G is not None and G > 0:
                for i in range(len(self.components), G):
                    new_comp = Component(name="Component %d" % (i + 1), parent=self)
                    self.components.append(new_comp)
                    self.observe_model(new_comp)

            # Observe components
            for component in self.components:
                self.observe_model(component)

            # Connect signals to lists and dicts:
            self._components_observer = ListObserver(
                self.on_component_inserted,
                self.on_component_removed,
                prop_name="components",
                model=self
            )

            self.probabilities = self.parse_init_arg(
                self.get_kwarg(kwargs, None, "probabilities", "data_probabilities"),
                get_correct_probability_model(R, G), default_is_class=True, child=True)
            self.probabilities.update() # force an update
            inherit_probabilities = kwargs.pop("inherit_probabilities", None)
            if inherit_probabilities is not None:
                for prop in self.probabilities.Meta.get_inheritable_properties():
                    setattr(self.probabilities, prop.inherit_flag, bool(inherit_probabilities))

            self._based_on_uuid = self.get_kwarg(kwargs, None, "based_on_uuid")
            self._based_on_index = self.get_kwarg(kwargs, None, "based_on_index")

    def __repr__(self):
        return "Phase(name='%s', based_on=%r)" % (self.name, self.based_on)

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    def on_component_inserted(self, item):
        # Set parent and observe the new component (visuals changed signals):
        if item.parent != self: item.parent = self
        self.observe_model(item)

    def on_component_removed(self, item):
        with self.data_changed.hold_and_emit():
            # Clear parent & stop observing:
            item.parent = None
            self.relieve_model(item)

    @Observer.observe("data_changed", signal=True)
    def notify_data_changed(self, model, prop_name, info):
        if isinstance(model, Phase) and model == self.based_on:
            with self.data_changed.hold():
                # make sure inherited probabilities are up-to-date
                self.probabilities.update()
                self.data_changed.emit(arg="based_on")
        else:
            self.data_changed.emit()

    @Observer.observe("visuals_changed", signal=True)
    def notify_visuals_changed(self, model, prop_name, info):
        self.visuals_changed.emit()

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def resolve_json_references(self):
        # Set the based on and linked with variables:
        if hasattr(self, "_based_on_uuid") and self._based_on_uuid is not None:
            self.based_on = type(type(self)).object_pool.get_object(self._based_on_uuid)
            del self._based_on_uuid
        elif hasattr(self, "_based_on_index") and self._based_on_index is not None and self._based_on_index != -1:
            warn("The use of object indices is deprecated since version 0.4. Please switch to using object UUIDs.", DeprecationWarning)
            self.based_on = self.parent.phases.get_user_from_index(self._based_on_index)
            del self._based_on_index
        for component in self.components:
            component.resolve_json_references()
        with self.data_changed.hold():
            # make sure inherited probabilities are up-to-date
            self.probabilities.update()

    def _pre_multi_save(self, phases, ordered_phases):
        ## Override from base class

        if self.based_on != "" and not self.based_on in phases:
            self.save_links = False
        Component.export_atom_types = True
        for component in self.components:
            component.save_links = self.save_links

        # Make sure parent is first in ordered list:
        if self.based_on in phases:
            index = ordered_phases.index(self)
            index2 = ordered_phases.index(self.based_on)
            if index < index2:
                ordered_phases.remove(self.based_on)
                ordered_phases.insert(index, self.based_on)

    def _post_multi_save(self):
        ## Override from base class
        self.save_links = True
        for component in self.components:
            component.save_links = True
        Component.export_atom_types = False

    def json_properties(self):
        retval = super(Phase, self).json_properties()
        if not self.save_links:
            for prop in self.Meta.all_properties:
                if getattr(prop, "inherit_flag", False):
                    retval[prop.inherit_flag] = False
            retval["based_on_uuid"] = ""
        else:
            retval["based_on_uuid"] = self.based_on.uuid if self.based_on else ""
        return retval

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def _update_interference_distributions(self):
        return self.CSDS_distribution.distrib

    def get_based_on_root(self):
        """
            Gets the root object in the based_on chain
        """
        if self.based_on is not None:
            return self.based_on.get_based_on_root()
        else:
            return self

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from math import sqrt

import sys
from operator import mul, truediv

import numpy as np

from deap import creator, base
from deap.tools import ParetoFront

class pyxrd_array(creator._numpy_array):
    """
        Helper DEAP.creator._numpy_array subclass that tracks changes to its
        underlying numerical data object, and if changed gets the corresponding
        data object from the context (if set). Allows for async evaluation of
        fitnesses for PyXRD parameter solutions.
    """
    min_bounds = None
    max_bounds = None

    def to_ndarray(self):
        return np.ndarray.copy(self)

    def __setitem__(self, i, y):
        y = min(y, self.max_bounds[i])
        y = max(y, self.min_bounds[i])
        creator._numpy_array.__setitem__(self, i, y)

    def __setslice__(self, i, j, y):
        y = np.array(y)
        np.clip(y, self.min_bounds[i:j], self.max_bounds[i:j], y)
        creator._numpy_array.__setslice__(self, i, j, y)

    pass # end of class

def result_func(individual, fitness):
    individual.fitness.values = fitness

class FitnessMin(base.Fitness):
    weights = []

    def getValues(self):
        return tuple(map(truediv, self.wvalues, (-1,) * len(self.wvalues)))

    def setValues(self, values):
        try:
            self.wvalues = tuple(map(mul, values, (-1,) * len(values)))
        except TypeError:
            _, _, traceback = sys.exc_info()
            raise TypeError("Both weights and assigned values must be a "
            "sequence of numbers when assigning to values of "
            "%r. Currently assigning value(s) %r of %r to a fitness with "
            "weights %s."
            % (self.__class__, values, type(values), (-1,) * len(values))).with_traceback(traceback)

    values = property(getValues, setValues, base.Fitness.delValues,
    ("Fitness values. Use directly ``individual.fitness.values = values`` "
     "in order to set the fitness and ``del individual.fitness.values`` "
     "in order to clear (invalidate) the fitness. The (unweighted) fitness "
     "can be directly accessed via ``individual.fitness.values``."))

    def __eq__(self, other):
        return tuple(self.wvalues) == tuple(other.wvalues)

    pass #end of class

class PyXRDParetoFront(ParetoFront):

    def get_best_n(self, n):
        inds = []
        for ind in self:
            d = 0.0
            for f in ind.fitness.wvalues:
                d += f ** 2
            d = sqrt(d)
            inds.append((d, ind))

        inds.sort(key=lambda ind:-ind[0])
        return inds[:n]

    def get_best(self):
        _, ind = self.get_best_n(1)[0]
        return ind

    def update(self, population):
        """Update the Pareto front hall of fame with the *population* by adding 
        the individuals from the population that are not dominated by the hall
        of fame. If any individual in the hall of fame is dominated it is
        removed.
        
        :param population: A list of individual with a fitness attribute to
                           update the hall of fame with.
        """
        for ind in population:
            is_dominated = False
            has_twin = False
            to_remove = []
            for i, hofer in enumerate(self):    # hofer = hall of famer
                try:
                    if hofer.fitness.dominates(ind.fitness):
                        is_dominated = True
                        break
                    elif ind.fitness.dominates(hofer.fitness):
                        to_remove.append(i)
                    elif ind.fitness == hofer.fitness and self.similar(ind, hofer):
                        has_twin = True
                        break
                except ValueError:
                    print(ind, ind.fitness)
                    print(hofer, hofer.fitness)
                    raise

            for i in reversed(to_remove):       # Remove the dominated hofer
                self.remove(i)
            if not is_dominated and not has_twin:
                self.insert(ind)
        # Run the garbage collector once for good measure
        import gc
        gc.collect()



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from distutils.version import LooseVersion

import json
import zipfile
import os, io

from shutil import move

from pyxrd.__version import __version__

from pyxrd.file_parsers.base_parser import BaseParser
from pyxrd.generic.io.json_codec import PyXRDDecoder, PyXRDEncoder
from pyxrd.generic.io.custom_io import storables, COMPRESSION

from pyxrd.generic.io.utils import unicode_open

class JSONParser(BaseParser):
    """
        PyXRD Object JSON Parser
    """

    description = "PyXRD Object JSON"
    mimetypes = ["application/octet-stream", "application/zip"]

    __file_mode__ = "r"

    @classmethod
    def _get_file(cls, fp, close=None):
        """
            Returns a three-tuple:
            filename, zipfile-object, close
        """
        if isinstance(fp, str): # fp is a filename
            filename = fp
            if zipfile.is_zipfile(filename):
                fp = zipfile.ZipFile(filename, cls.__file_mode__)
            else:
                fp = unicode_open(filename, cls.__file_mode__)
            close = True if close is None else close
        else: # fp is a file pointer
            filename = getattr(fp, 'name', None)
            if zipfile.is_zipfile(fp):
                fp = zipfile.ZipFile(fp)
            close = False if close is None else close
        return filename, fp, close

    @classmethod
    def _parse_header(cls, filename, fp, data_objects=None, close=False):
        return data_objects # just pass it on, nothing to do

    @classmethod
    def _parse_data(cls, filename, fp, data_objects=None, close=True):
        # At this point filename is just there for information; fp can safely be
        # assumed to be a file pointer - if not, not our problem
        is_zipfile = isinstance(fp, zipfile.ZipFile)
        if is_zipfile: # ZIP files
                namelist = fp.namelist()
                if 'content' in namelist: # Multi-part object (e.g. project)
                    obj = None
                    decoder = json.JSONDecoder()

                    def get_named_item(fpt, name):
                        try:
                            cf = fpt.open(name, cls.__file_mode__)
                            obj = decoder.decode(cf.read().decode("utf-8"))
                        finally:
                            cf.close()
                        return obj
                    
                    # Parse the content file
                    obj = get_named_item(fp, 'content')
                    
                    # Check for a version tag:
                    if 'version' in namelist:
                        namelist.remove('version')
                        version = get_named_item(fp, 'version')
                        if LooseVersion(version) > LooseVersion(__version__.replace("v", "")):
                            raise RuntimeError("Unsupported project" + \
                                  "version '%s', program version is '%s'" % (
                                version, __version__
                            ))
                    else:
                        logging.warn("Loading pre-v0.8 file format, " +
                                     "might be deprecated!")

                    # Make sure we have a dict at this point
                    if not hasattr(obj, "update"):
                        raise RuntimeError("Decoding a multi-part JSON " + \
                          "object requires the root to be a dictionary object!")

                    # Parse all the other files, and set accordingly in the content dict
                    for sub_name in namelist:
                        if sub_name != "content":
                            obj["properties"][sub_name] = get_named_item(
                                fp, sub_name)

                    # Now parse the JSON dict to a Python object
                    data_objects = PyXRDDecoder(mapper=storables).__pyxrd_decode__(obj) or obj
                else: # Multiple objects in one zip file (e.g. phases)
                    data_objects = []
                    for sub_name in namelist:
                        zpf = fp.open(sub_name, cls.__file_mode__)
                        data_objects.append(cls.parse(zpf))
                        zpf.close()
        elif hasattr(fp, 'seek'): # Regular file
            try:
                fp.seek(0) # reset file position
            except io.UnsupportedOperation:
                pass # ignore these
            data_objects = PyXRDDecoder.decode_file(fp, mapper=storables)
        else:
            pass # use filename?

        if close: fp.close()
        return data_objects

    @staticmethod
    def write(obj, file, zipped=False): # @ReservedAssignment
        """
        Saves the output from dump_object() to `filename`, optionally zipping it.
        File can be either a filename or a file-like object. If it is a filename
        extra precautions are taken to prevent malformed data overwriting a good
        file. With file objects this is not the case.
        """
        filename = None
        if isinstance(file, str):
            # We have a filename, not a file object
            filename = file
            # Create temporary filenames for output, and a backup filename if
            # the file already exists.
            file = filename + ".tmp" # @ReservedAssignment
            backup_name = filename + "~"
        try:
            if zipped:
                # Try to safe the file as a zipfile:
                with zipfile.ZipFile(file, mode="w", compression=COMPRESSION) as f:
                    for partname, json_object in obj.to_json_multi_part():
                        f.writestr(partname, PyXRDEncoder.dump_object(json_object))
            else:
                # Regular text file:
                if filename is not None:
                    with unicode_open(file, 'w') as f:
                        PyXRDEncoder.dump_object_to_file(obj, f)
                else:
                    PyXRDEncoder.dump_object_to_file(obj, file)
        except:
            # In case saving fails, remove the temporary file:
            if filename is not None and os.path.exists(file):
                os.remove(file)
            raise

        if filename is not None:
            # If target file exists, back it up:
            if os.path.exists(filename):
                move(filename, backup_name)
            # Rename temporary saved file:
            move(file, filename)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory

from pyxrd.generic.views.treeview_tools import new_text_column, new_combo_column, create_float_data_func, setup_treeview
from pyxrd.generic.controllers.objectliststore_controllers import wrap_list_property_to_treemodel
from pyxrd.generic.controllers import InlineObjectListStoreController

from pyxrd.atoms.models import Atom
from pyxrd.project.models import Project

class EditLayerController(InlineObjectListStoreController):
    """ 
        Controller for the (inter)layer atom ObjectListStores
    """
    auto_adapt = False
    enable_import = True
    enable_export = True
    treemodel_class_type = Atom
    file_filters = Atom.Meta.layer_filters
    new_atom_type = None

    @property
    def atom_types_treemodel(self):
        return wrap_list_property_to_treemodel(
            self.model.phase.project, Project.atom_types)

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def _setup_treeview(self, tv, model):
        setup_treeview(tv, model, sel_mode='MULTIPLE', reset=True)
        tv.set_model(model)

        # Add Atom name, default z, calculated z and #:
        def add_text_col(title, colnr, is_float=True, editable=True):
            tv.append_column(new_text_column(
                title,
                data_func=create_float_data_func() if is_float else None,
                editable=editable,
                edited_callback=(self.on_item_cell_edited, (model, colnr)) if editable else None,
                resizable=True,
                text_col=colnr))
        add_text_col('Atom name', model.c_name, is_float=False)
        add_text_col('Def. Z (nm)', model.c_default_z)
        add_text_col('Calc. Z (nm)', model.c_z, editable=False)
        add_text_col('#', model.c_pn)

        # Add atom type column (combo box with atom types from pyxrd.project level):
        def atom_type_renderer(column, cell, model, itr, col=None):
            try:
                name = model.get_user_data_from_path(model.get_path(itr)).atom_type.name
            except:
                name = '#NA#'
            cell.set_property('text', name)
            return
        def adjust_combo(cell, editable, path, data=None):
            editable.set_wrap_width(10)
        tv.append_column(new_combo_column(
            "Element",
            data_func=(atom_type_renderer, (3,)),
            changed_callback=self.on_atom_type_changed,
            edited_callback=(self.on_atom_type_edited, (model,)),
            editing_started_callback=adjust_combo,
            model=self.atom_types_treemodel,
            text_column=self.atom_types_treemodel.c_name,
            editable=True,
            has_entry=True))

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def create_new_object_proxy(self):
        return Atom(name="New Atom", parent=self.model)

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_save_object_clicked(self, widget, user_data=None):
        def on_accept(save_dialog):
            Atom.save_as_csv(save_dialog.filename, self.get_all_objects())
        current_name = "%s%s" % (
            self.model.name.lower(),
            self.treemodel_property_name.replace("data", "").lower()
        )
        DialogFactory.get_save_dialog(
            "Export atoms", parent=self.view.get_toplevel(),
            current_name=current_name, filters=self.file_filters,
        ).run(on_accept)

    def on_load_object_clicked(self, widget, user_data=None):
        def import_layer(dialog):
            def on_accept(dialog):
                del self.treemodel_data[:] # clears the list
                Atom.get_from_csv(dialog.filename, self.treemodel_data.append, self.model)
            DialogFactory.get_load_dialog(
                "Import atoms", parent=self.view.get_toplevel(),
                filters=self.file_filters
            ).run(on_accept)
        DialogFactory.get_confirmation_dialog(
            message="Are you sure?\nImporting a layer file will clear the current list of atoms!",
            parent=self.view.get_toplevel()
        ).run(import_layer)

    def on_atom_type_changed(self, combo, path, new_iter, user_data=None):
        """Called when the user selects an AtomType from the combo box"""
        self.new_atom_type = self.atom_types_treemodel.get_user_data(new_iter)
        return True

    def on_atom_type_edited(self, combo, path, new_text, user_data=None):
        """Called when the user has closed the AtomType combo box 
        (so after the on_atom_type_changed call)"""
        atom = self.treemodel_data[int(path)]
        if atom is not None:
            # If new_atom_type is not set, but the user has typed in the name
            # of an atom_type, find it (index search):
            if self.new_atom_type is None and not new_text in (None, ""):
                for atom_type in self.model.phase.project.atom_types:
                    if atom_type.name == new_text:
                        self.new_atom_type = atom_type
            # Set the new atom type if it is not None:
            if self.new_atom_type is not None:
                atom.atom_type = self.new_atom_type
            # Clear variable and leave
            self.new_atom_type = None
            return True
        return False

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from random import choice
import numpy as np

from mvc.models.properties import LabeledProperty
from mvc.models.properties.signal_mixin import SignalMixin
from mvc.models.properties.observe_mixin import ObserveMixin

from pyxrd.generic.io import storables, get_case_insensitive_glob
from pyxrd.generic.models.lines import PyXRDLine
from pyxrd.refinement.refinables.metaclasses import PyXRDRefinableMeta
from pyxrd.refinement.refinables.mixins import RefinementGroup
from pyxrd.file_parsers.xrd_parsers import xrd_parsers

from .abstract_phase import AbstractPhase

@storables.register()
class RawPatternPhase(RefinementGroup, AbstractPhase, metaclass=PyXRDRefinableMeta):

    # MODEL INTEL:
    class Meta(AbstractPhase.Meta):
        store_id = "RawPatternPhase"
        file_filters = [
            ("Phase file", get_case_insensitive_glob("*.PHS")),
        ]
        rp_filters = xrd_parsers.get_import_file_filters()
        rp_export_filters = xrd_parsers.get_export_file_filters()

    _data_object = None
    @property
    def data_object(self):
        self._data_object.type = "RawPatternPhase"

        self._data_object.raw_pattern_x = self.raw_pattern.data_x
        self._data_object.raw_pattern_y = self.raw_pattern.data_y[:, 0]
        self._data_object.apply_lpf = False
        self._data_object.apply_correction = False

        return self._data_object

    project = property(AbstractPhase.parent.fget, AbstractPhase.parent.fset)

    @property
    def refine_title(self):
        return "Raw Pattern Phase"

    @property
    def is_refinable(self):
        return False

    @property
    def children_refinable(self):
        return False

    @property
    def refinables(self):
        return []

    raw_pattern = LabeledProperty(
        default=None, text="Raw pattern",
        visible=True, persistent=True, tabular=True,
        inheritable=True, inherit_flag="inherit_CSDS_distribution", inherit_from="based_on.CSDS_distribution",
        signal_name="data_changed",
        mix_with=(SignalMixin, ObserveMixin,)
    )

    @property
    def spec_max_display_y(self):
        """The maximum intensity (y-axis) of the current loaded profile"""
        _max = 0.0
        if self.raw_pattern is not None:
            _max = max(_max, np.max(self.raw_pattern.max_display_y))
        return _max

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        my_kwargs = self.pop_kwargs(kwargs,
            *[prop.label for prop in RawPatternPhase.Meta.get_local_persistent_properties()]
        )
        super(RawPatternPhase, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        with self.data_changed.hold():
            self.raw_pattern = PyXRDLine(
                data=self.get_kwarg(kwargs, None, "raw_pattern"),
                parent=self
            )
            self.display_color = self.get_kwarg(kwargs, choice(self.line_colors), "display_color")
            self.inherit_display_color = self.get_kwarg(kwargs, False, "inherit_display_color")

    def __repr__(self):
        return "RawPatternPhase(name='%s')" % (self.name)

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @AbstractPhase.observe("data_changed", signal=True)
    def notify_data_changed(self, model, prop_name, info):
        self.data_changed.emit() # propagate signal

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

class _RefinementBase(object):
    """
    Base class for `RefinementGroup` and `RefinementValue` mixins. It's 
    used to provide common functionality and a way to check for the kind of
    refinement class we're dealing with when building the refinement tree.
            
    .. attribute:: refine_title

        A string used as the title for the group in the refinement tree

    .. attribute:: refine_descriptor

        A longer title string which gives more information (phase, component, etc) 
        
    .. attribute:: is_refinable

        Whether or not this instance is refinable
        
    .. attribute:: refinables
        
        An iterable with the names of the refinable properties 
        
    .. attribute:: refine_value
    
        Mapper for the actual refinable value (if available). This should be
        overriden by deriving classes.
        
    """

    @property
    def refine_title(self):
        return "Refinement Base"

    @property
    def refine_descriptor_data(self):
        return dict()

    @property
    def is_refinable(self):
        return True

    @property
    def refinables(self):
        return []

    @property
    def refine_info(self):
        return None

    @property
    def refine_value(self):
        return None
    @refine_value.setter
    def refine_value(self, value):
        pass

    pass # end of class

class RefinementGroup(_RefinementBase):
    """
    Mixin for objects that are not refinable themselves,
    but have refinable properties. They are presented in the refinement
    tree using their title value.
    Subclasses should override refine_title to make it more descriptive.
    
    .. attribute:: children_refinable

        Whether or not the child properties of this group can be refinable.
        This should normally always be True, unless for example if the entire
        group of properties have a single inherit property.
    
    """

    @property
    def refine_title(self):
        return "Refinement Group"

    @property
    def is_refinable(self):
        return False

    @property
    def children_refinable(self):
        return True

    @property
    def refinables(self):
        return self.Meta.get_refinable_properties()

    pass # end of class

class RefinementValue(_RefinementBase):
    """
        Mixin for objects that hold a single refinable property. They are
        collapsed into one line in the refinement tree. 
        Subclasses should override both the refine_title property to make it
        more descriptive, and the refine_value property to return and set the
        correct (refinable) attribute.
    """

    @property
    def refine_title(self):
        return "Refinement Value"

    pass # end of class



#!/usr/bin/python
# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from io import StringIO

from test.test_generic.test_io.test_file_parsers import BaseTestParsers
from pyxrd.file_parsers.xrd_parsers import UDFParser


__all__ = [
    'TestUDFParser',
]

class TestUDFParser(BaseTestParsers.BaseTestParser):

    parser_class = UDFParser
    file_data = [
        StringIO(r"""SampleIdent,Sample5 ,/
Title1,Dat2rit program ,/
Title2,Sample5 ,/
DataAngleRange,   5.0000, 5.6400,/
ScanStepSize,     0.020,/
RawScan
    8000,    7000,    6000,    5000,    4000,    3000,    2000,    1000
    800,     700,     600,     500,     400,     300,     200,     100
    80,      70,      60,      50,      40,      30,      20,      10
    8,       7,       6,       5,       4,       3,       2,       1
    0/"""),
    ]

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

import numpy as np

from pyxrd.calculations.data_objects import CSDSData
from pyxrd.calculations.CSDS import calculate_distribution

__all__ = [
    'TestCSDSCalcs',
]

class TestCSDSCalcs(unittest.TestCase):

    def setUp(self):
        self.CSDS_data_kwargs = dict(
            average = 10,
            maximum = 50,
            minimum = 1,
            alpha_scale = 0.9485,
            alpha_offset = 0.017,
            beta_scale = 0.1032,
            beta_offset = 0.0034
        )
        self.CSDS_data = CSDSData(**self.CSDS_data_kwargs)

    def tearDown(self):
        del self.CSDS_data

    def test_not_none(self):
        self.assertIsNotNone(self.CSDS_data)
        
    def test_attributes(self):
        for key, value in self.CSDS_data_kwargs.items():
            self.assertEquals(getattr(self.CSDS_data, key), value)

    def test_calculate_distribution(self):
        result = calculate_distribution(
            self.CSDS_data
        )
        self.assertIsNotNone(result)
        self.assertEquals(len(result), 2)

    pass # end of class


calculate_distribution


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc.models.properties import FloatProperty, BoolProperty

from pyxrd.generic.models.base import PyXRDModel
from pyxrd.generic.io import storables, Storable
from pyxrd.generic.utils import not_none

@storables.register()
class RefinementInfo(PyXRDModel, Storable):
    """
        A model that is used to store the refinement information for each
        refinable value (in other models): minimum and maximum value and
        a flag to indicate whether this value is selected for refinement.
    """

    # MODEL INTEL:
    class Meta(PyXRDModel.Meta, Storable.Meta):
        store_id = "RefinementInfo"

    minimum = FloatProperty(default=0.0, text="Minimum", persistent=True)
    maximum = FloatProperty(default=1.0, text="Maximum", persistent=True)
    refine = BoolProperty(default=False, text="Refine", persistent=True)

    def __init__(self, minimum, maximum, refine, *args, **kwargs):
        """
            Valid *positional* arguments for a RefinementInfo are:
                refine: whether or not the linked parameter is selected for refinement
                minimum: the minimum allowable value for the linked parameter
                maximum: the maximum allowable value for the linked parameter   
        """
        super(RefinementInfo, self).__init__()
        self.refine = refine
        self.minimum = not_none(minimum, 0.0)
        self.maximum = not_none(maximum, 1.0)

    def to_json(self):
        return self.json_properties()

    def json_properties(self):
        return [self.minimum, self.maximum, self.refine]

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from contextlib import contextmanager

from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory
from mvc import Controller

from pyxrd.generic.plot.eye_dropper import EyeDropper
from pyxrd.generic.plot.draggables import DraggableVLine
from pyxrd.generic.controllers import DialogController, BaseController, ObjectListStoreController
from pyxrd.generic.views.treeview_tools import setup_treeview, new_text_column, new_toggle_column
from pyxrd.generic.io.utils import get_case_insensitive_glob

from pyxrd.specimen.models import Marker, ThresholdSelector, MineralScorer
from pyxrd.specimen.views import (
    EditMarkerView,
    DetectPeaksView,
    MatchMineralsView
)
from pyxrd.generic.utils import not_none
from pyxrd.data import settings

class EditMarkerController(BaseController):

    def register_view(self, view):
        self.update_sensitivities()
        self.update_nanometer()

    def update_sensitivities(self):
        for name in ("style", "base", "align", "top", "color"):
            wid = self.view["marker_%s" % name]
            wid.set_sensitive(not getattr(self.model, "inherit_%s" % name)) 
        for name in ("angle", "top_offset"):
            self.view["spb_%s" % name].set_sensitive(not getattr(self.model, "inherit_%s" % name))

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Controller.observe("position", assign=True, after=True)
    def notif_parameter_changed(self, model, prop_name, info):
        if prop_name == "position":
            self.update_nanometer()

    @Controller.observe("inherit_style", assign=True)
    @Controller.observe("inherit_align", assign=True)
    @Controller.observe("inherit_base", assign=True)
    @Controller.observe("inherit_top", assign=True)
    @Controller.observe("inherit_top_offset", assign=True)
    @Controller.observe("inherit_angle", assign=True)
    @Controller.observe("inherit_color", assign=True)
    def notif_angle_toggled(self, model, prop_name, info):
        self.update_sensitivities()

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_nanometer_changed(self, widget):
        try:
            position = float(widget.get_value())
        except ValueError:
            logger.exception("User set nanometers to an invalid value: %s", widget.get_value())
        else:
            self.model.set_nm_position(position)
        pass

    def update_nanometer(self):
        self.view["entry_nanometer"].set_value(self.model.get_nm_position())

    def on_sample_clicked(self, widget):

        def click_callback(x_pos, event):
            if self.edc is not None:
                self.edc.disconnect()
            self.view.get_toplevel().present()
            if x_pos != -1:
                self.model.position = x_pos

        self.edc = EyeDropper(
            self.parent.plot_controller,
            click_callback
        )

        self.view.get_toplevel().hide()
        self.parent.view.get_toplevel().present()

class MarkersController(ObjectListStoreController):
    """ 
        Controller for the markers list
    """

    file_filters = ("Marker file", get_case_insensitive_glob("*.MRK")),
    treemodel_property_name = "markers"
    treemodel_class_type = Marker
    columns = [
        (" ", "c_visible"),
        ("Marker label", "c_label")
    ]
    delete_msg = "Deleting a marker is irreversible!\nAre You sure you want to continue?"
    obj_type_map = [
        (Marker, EditMarkerView, EditMarkerController),
    ]
    title = "Edit Markers"

    def get_markers_tree_model(self, *args):
        return self.treemodel

    def setup_treeview_col_c_visible(self, treeview, name, col_descr, col_index, tv_col_nr):
        def toggle_renderer(column, cell, model, itr, data=None):
            try:
                col = column.get_col_attr("active")
                value = model.get_value(itr, col)
                cell.set_property('active', not_none(value, False))
            except TypeError:
                if settings.DEBUG: raise
                pass

        col = new_toggle_column(" ",
                toggled_callback=(self.on_marker_visible_toggled, (treeview.get_model(), col_index)),
                data_func=toggle_renderer,
                resizable=False,
                expand=False,
                activatable=True,
                active_col=col_index)
        setattr(col, "colnr", col_index)
        treeview.append_column(col)
        return True

    def select_markers(self, markers):
        self.set_selected_objects()

    @contextmanager
    def _multi_operation_context(self):
        with self.model.visuals_changed.hold():
            yield

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_load_object_clicked(self, event):
        def on_accept(dialog):
            with self._multi_operation_context():
                for marker in Marker.get_from_csv(dialog.filename, self.model):
                    self.model.markers.append(marker)
        DialogFactory.get_load_dialog(
            "Import markers", parent=self.view.get_top_widget(),
            filters=self.file_filters
        ).run(on_accept)

    def on_save_object_clicked(self, event):
        def on_accept(dialog):
            Marker.save_as_csv(dialog.filename, self.get_selected_objects())
        DialogFactory.get_save_dialog(
            "Export markers", parent=self.view.get_top_widget(),
            filters=self.file_filters
        ).run(on_accept)

    def create_new_object_proxy(self):
        return Marker(label="New Marker", parent=self.model)

    def on_marker_visible_toggled(self, cell, path, model, colnr):
        if model is not None:
            itr = model.get_iter(path)
            model.set_value(itr, colnr, not cell.get_active())
            return True
        return False

    @BaseController.status_message("Finding peaks...", "find_peaks")
    def on_find_peaks_clicked(self, widget):
        def after_cb(threshold):
            self.model.auto_add_peaks(threshold)

        sel_model = ThresholdSelector(parent=self.model)
        sel_view = DetectPeaksView(parent=self.view)
        sel_ctrl = ThresholdController(model=sel_model, view=sel_view, parent=self, callback=after_cb) #@UnusedVariable
        
        show_threshold_plot = DialogFactory.get_progress_dialog(
            action=sel_model.update_threshold_plot_data,
            complete_callback=lambda *a, **k: sel_view.present(),
            gui_message="Finding peaks {progress:.0f}%...",
            toplevel=self.view.get_top_widget()
        )

        if len(self.model.markers) > 0:
            def on_accept(dialog):
                self.model.clear_markers()
                show_threshold_plot()
            def on_reject(dialog):
                show_threshold_plot()
            DialogFactory.get_confirmation_dialog(
                "Do you want to clear the current markers for this pattern?",
                parent=self.view.get_top_widget()
            ).run(on_accept, on_reject)
        else:
            show_threshold_plot()          

    def on_match_minerals_clicked(self, widget):
        def apply_cb(matches):
            with self._multi_operation_context():
                for name, abbreviation, peaks, matches, score in matches: #@UnusedVariable
                    for marker in self.get_selected_objects():
                        for mpos, epos in matches: #@UnusedVariable
                            if marker.get_nm_position() * 10. == epos:
                                marker.label += ", %s" % abbreviation

        def close_cb():
            self.model.visuals_changed.emit()
            self.view.show()

        marker_peaks = [] # position, intensity

        for marker in self.get_selected_objects():
            intensity = self.model.experimental_pattern.get_y_at_x(
                marker.position)
            marker_peaks.append((marker.get_nm_position() * 10., intensity))

        scorer_model = MineralScorer(marker_peaks=marker_peaks, parent=self.model)
        scorer_view = MatchMineralsView(parent=self.view)
        scorer_ctrl = MatchMineralController(model=scorer_model, view=scorer_view, parent=self, apply_callback=apply_cb, close_callback=close_cb) #@UnusedVariable

        self.view.hide()
        scorer_view.present()

    pass # end of class

class MatchMineralController(DialogController):

    apply_callback = None
    close_callback = None

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, model, view, spurious=False, auto_adapt=False, parent=None, apply_callback=None, close_callback=None):
        DialogController.__init__(self, model=model, view=view, spurious=spurious, auto_adapt=auto_adapt, parent=parent)
        self.apply_callback = apply_callback
        self.close_callback = close_callback

    def register_adapters(self):
        super(MatchMineralController, self).register_adapters()
        if self.model is not None:
            self.reload_minerals()
            self.reload_matches()

    def register_view(self, view):
        super(MatchMineralController, self).register_view(view)
        if view is not None:
            # MATCHES Treeview:
            tv = self.view['tv_matches']

            setup_treeview(tv, None,
                reset=True,
                on_selection_changed=self.selection_changed,
            )

            tv.append_column(new_text_column(
                "Name", markup_col=0,
                xalign=0,
            ))

            tv.append_column(new_text_column(
                "Abbr.", markup_col=1,
                expand=False,
            ))

            def get_value(column, cell, model, itr, *args):
                value = model.get_value(itr, column.get_col_attr('markup'))
                try: value = "%.5f" % value
                except TypeError: value = ""
                cell.set_property("markup", value)
                return
            tv.append_column(new_text_column(
                "Score",
                markup_col=4,
                expand=False,
                data_func=get_value
            ))

            # ALL MINERALS Treeview:
            tv = self.view['tv_minerals']
            setup_treeview(tv, None,
                reset=True,
                on_selection_changed=self.selection_changed,
            )

            tv.append_column(new_text_column(
                "Name", markup_col=0,
                xalign=0,
            ))

            tv.append_column(new_text_column(
                "Abbr.", markup_col=1,
                expand=False,
            ))

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Controller.observe("matches_changed", signal=True)
    def notif_parameter_changed(self, model, prop_name, info):
        self.reload_matches()

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def selection_changed(self, selection, *args):
        if selection.count_selected_rows() >= 1:
            model, paths = selection.get_selected_rows()
            itr = model.get_iter(paths[0])
            name, _, peaks = model.get(itr, 0, 1, 2)
            self.model.specimen.mineral_preview = (name, peaks)
            self.model.specimen.visuals_changed.emit()

    def on_auto_match_clicked(self, event):
        self.model.auto_match()

    def on_add_match_clicked(self, event):
        selection = self.view.tv_minerals.get_selection()
        if selection.count_selected_rows() >= 1:
            model, paths = selection.get_selected_rows()
            itr = model.get_iter(paths[0])
            name, abbreviation, peaks = model.get(itr, 0, 1, 2)
            self.model.add_match(name, abbreviation, peaks)

    def on_del_match_clicked(self, event):
        selection = self.view.tv_matches.get_selection()
        if selection.count_selected_rows() >= 1:
            _, paths = selection.get_selected_rows()
            self.model.del_match(*paths[0])

    def on_apply_clicked(self, event):
        if self.apply_callback is not None and callable(self.apply_callback):
            self.model.specimen.mineral_preview = None
            self.apply_callback(self.model.matches)
        self.view.hide()

    def on_cancel(self):
        if self.close_callback is not None and callable(self.close_callback):
            self.model.specimen.mineral_preview = None
            self.close_callback()
        self.view.hide()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def reload_matches(self):
        if not hasattr(self, 'tv_matches_model'):
            self.tv_matches_model = Gtk.ListStore(str, str, object, object, float)
        else:
            self.tv_matches_model.clear()
        for name, abbreviation, peaks, matches, score in self.model.matches:
            self.tv_matches_model.append([name, abbreviation, peaks, matches, score])

        tv = self.view.tv_matches
        tv.set_model(self.tv_matches_model)

    def reload_minerals(self):
        if not hasattr(self, 'tv_matches_model'):
            self.tv_minerals_model = Gtk.ListStore(str, str, object)
        else:
            self.tv_minerals_model.clear()
        for name, abbreviation, peaks in self.model.minerals:
            self.tv_minerals_model.append([name, abbreviation, peaks])

        tv = self.view.tv_minerals
        tv.set_model(self.tv_minerals_model)

    pass # end of class


class ThresholdController(DialogController):

    auto_adapt_included = [
        "pattern",
        "sel_threshold",
        "max_threshold",
        "sel_num_peaks",
        "steps"
    ]

    callback = None
    dline = None

    def __init__(self, *args, **kwargs):
        callback = kwargs.pop("callback", None)
        super(ThresholdController, self).__init__(*args, **kwargs)
        self.callback = callback
        self.dline = None

    def update_plot(self):
        if self.view is not None:
            self.view.plot.cla()
            if self.dline is not None:
                self.dline.disconnect()
                self.dline = None

            def dline_cb(x):
                self.model.sel_threshold = x

            if self.model is not None and self.model.threshold_plot_data is not None:
                x, y = self.model.threshold_plot_data
                self.view.plot.plot(x, y, 'k-')
                self.line = self.view.plot.axvline(x=self.model.sel_threshold, color="#0000FF", linestyle="-")
                self.dline = DraggableVLine(self.line, callback=dline_cb, window=self.view.matlib_canvas.get_window())
            self.view.plot.set_ylabel('# of peaks', labelpad=1)
            self.view.plot.set_xlabel('Threshold', labelpad=1)
            self.view.figure.subplots_adjust(left=0.15, right=0.875, top=0.875, bottom=0.15)
            self.view.plot.autoscale_view()
            self.view.matlib_canvas.draw()

    def register_adapters(self):
        self.update_plot()

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Controller.observe("sel_threshold", assign=True)
    @Controller.observe("threshold_plot_data", assign=True)
    def notif_parameter_changed(self, model, prop_name, info):
        self.update_plot()

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_btn_ok_clicked(self, event):
        if self.callback is not None and callable(self.callback):
            self.callback(self.model)
        return DialogController.on_btn_ok_clicked(self, event)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
from mvc.models.properties.tools import modify
logger = logging.getLogger(__name__)

from mvc.models.properties import (
    ListProperty, SignalMixin
)

from pyxrd.data import settings
from pyxrd.generic.io import storables

from pyxrd.generic.models.base import DataModel

from .pyxrd_line import PyXRDLine

@storables.register()
class CalculatedLine(PyXRDLine):

    # MODEL INTEL:
    class Meta(PyXRDLine.Meta):
        store_id = "CalculatedLine"
        inherit_format = "display_calc_%s"

    specimen = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:

    phase_colors = ListProperty(
        default=[], test="Phase colors",
        mix_with=(SignalMixin,),
        signal_name="visuals_changed",
    )

    #: The line color
    color = modify(PyXRDLine.color,
        default=settings.CALCULATED_COLOR,
        inherit_from="parent.parent.display_calc_color"
    )

    #: The linewidth in points
    lw = modify(PyXRDLine.lw,
        default=settings.CALCULATED_LINEWIDTH,
        inherit_from="parent.parent.display_calc_lw"
    )

    #: A short string describing the (matplotlib) linestyle
    ls = modify(PyXRDLine.ls,
        default=settings.CALCULATED_LINESTYLE,
        inherit_from="parent.parent.display_calc_ls"
    )

    #: A short string describing the (matplotlib) marker
    marker = modify(PyXRDLine.marker,
        default=settings.CALCULATED_MARKER,
        inherit_from="parent.parent.display_calc_marker"
    )

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np

from scipy.special import erf
from math import sqrt

from .math_tools import sqrt2pi, sqrt8

def get_S(soller1, soller2):
    _S = sqrt((soller1 * 0.5) ** 2 + (soller2 * 0.5) ** 2)
    _S1S2 = soller1 * soller2
    return _S, _S1S2

def get_T(range_theta, sigma_star, soller1, soller2):
    sigma_star = float(max(sigma_star, 1e-18))
    S, _ = get_S(soller1, soller2)
    range_st = np.sin(range_theta)
    Q = S / (sqrt8 * range_st * sigma_star)
    return erf(Q) * sqrt2pi / (2.0 * sigma_star * S) - 2.0 * range_st * (1.0 - np.exp(-(Q ** 2.0))) / (S ** 2.0)
 
def get_lorentz_polarisation_factor(range_theta, sigma_star, soller1, soller2, mcr_2theta):
    """
        Get the lorentz polarisation factor for the given sigma-star value,
        soller slits, monochromator Bragg angle and the given theta range
    """
    T = get_T(range_theta, sigma_star, soller1, soller2)
    pol = np.cos(np.radians(mcr_2theta)) ** 2
    return T * (1.0 + pol * (np.cos(2.0 * range_theta) ** 2)) / np.sin(range_theta)

def get_fixed_to_ads_correction_range(range_theta, goniometer):
    return np.sin(range_theta)

def get_nm_from_t(theta, wavelength=0.154056, zero_for_inf=False):
    """
        Convert the given theta angles (scalar or np array) to
        nanometer spacings using the given wavelength
    """
    return get_nm_from_2t(2.0 * theta, wavelength=wavelength, zero_for_inf=zero_for_inf)

def get_nm_from_2t(twotheta, wavelength=0.154056, zero_for_inf=False):
    """
        Convert the given 2-theta angles (scalar or np array) to
        nanometer spacings using the given wavelength
    """
    if twotheta == 0:
        return 0. if zero_for_inf else 1e16
    else:
        return wavelength / (2.0 * np.sin(np.radians(twotheta / 2.0)))

def get_t_from_nm(nm, wavelength=0.154056):
    """
        Convert the given nanometer spacings (scalar or np array) to
        theta angles using the given wavelength
    """
    return get_2t_from_nm(nm, wavelength=wavelength) / 2

def get_2t_from_nm(nm, wavelength=0.154056):
    """
        Convert the given nanometer spacings (scalar or np array) to
        2-theta angles using the given wavelength
    """
    twotheta = 0.0
    if nm != 0:
        twotheta = np.degrees(np.arcsin(max(-1.0, min(1.0, wavelength / (2.0 * nm))))) * 2.0
    return twotheta




#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np
from scipy.optimize import fmin_l_bfgs_b as optim



asf1 = np.array([0.126842, 4.708971, 1.194814, 1.558157, 1.170413, 3.239403, 4.875207, 108.506081, 0.111516, 48.292408, 1.928171])
asf2 = np.array([0.058851, 3.062918, 4.135106, 0.853742, 1.036792, 0.85252, 2.015803, 4.417941, 0.065307, 9.66971, 0.187818])

def func(asf, stl_range):
    f = np.zeros(stl_range.shape)
    for i in range(1, 6):
        f += asf[i] * np.exp(-asf[5 + i] * (stl_range) ** 2)
    f += asf[0]
    return f

stl_range = np.arange(0, 1, 0.01)

expected1 = func(asf1, stl_range)
expected2 = func(asf2, stl_range)

expected = (expected1 + expected2) / 2.0

def calculate_R2(x0, *args):
    global stl_range
    global expected
    calculated = func(x0, stl_range)
    return np.sum((calculated - expected) ** 2)

bounds = [
    (0, None),

    (None, None),
    (None, None),
    (None, None),
    (None, None),
    (None, None),

    (0.001, None),
    (0.001, None),
    (0.001, None),
    (0.001, None),
    (0.001, None),
]

x0 = asf2

lastx, lastR2, info = optim(calculate_R2, x0, approx_grad=True, pgtol=10e-24 , factr=2, iprint=-1, bounds=bounds)

print(lastR2)
print("\t".join([("%.10g" % fl).replace(".", ",") for fl in lastx]))
print(info)





# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import functools


from pyxrd.generic.asynchronous.cancellable import Cancellable
from pyxrd.generic.asynchronous.has_async_calls import HasAsyncCalls

class AsyncEvaluatable(HasAsyncCalls, Cancellable):
    
    # self.refiner.get_pickled_data_object_for_solution

    def do_async_evaluation(self, iter_func, eval_func, data_func, result_func):
        """ 
            Utility that combines a submit and fetch cycle in a single function
            call.
            iter_func is a generation callback (generates solutions)
            data_func transforms the given solutions to something eval_func can 
              work with (this can be a pass-through operation)
            eval_func evaluates a single (generated) solution (this must be
              picklable)
            result_func receives each solution and its residual as arguments
            
        """
        assert callable(iter_func)
        assert callable(eval_func)
        assert callable(data_func)
        assert callable(result_func)
        
        results = []
        solutions = []
        for solution in iter_func():
            result = self.submit_async_call(functools.partial(
                eval_func, data_func(solution)
            ))
            solutions.append(solution)
            results.append(result)
            if self._user_cancelled(): # Stop submitting new individuals
                break
            
        for solution, result in zip(solutions, map(self.fetch_async_result, results)): 
            result_func(solution, result)
        
        del results
        
        # Run the garbage collector once for good measure
        import gc
        gc.collect()
        
        return solutions


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import sys, os
import logging
logger = logging.getLogger(__name__)

import multiprocessing

def _worker_initializer(pool_stop, debug, *args):
    # Spoof command line arguments so settings are loaded with correct
    # debugging flag
    if debug and not "-d" in sys.argv:
        sys.argv.insert(1, "-d")
    if not debug and "-d" in sys.argv:
        sys.argv.remove("-d")

    # Load settings
    from pyxrd.data import settings
    settings.initialize()

    if settings.DEBUG:
        from pyxrd import stacktracer
        stacktracer.trace_start(
            "trace-worker-%s.html" % multiprocessing.current_process().name,
            interval=5, auto=True) # Set auto flag to always update file!
    logger.info("Worker process initialized, DEBUG=%s" % debug)
    pass

class PyXRDServer(object):

    pool = None
    pool_stop = None

    running = True

    def loopCondition(self):
        return self.running

    def __init__(self):
        from pyxrd.data import settings
        settings.initialize()

        logger.warning("Creating pool, DEBUG=%s" % settings.DEBUG)

        self.pool_stop = multiprocessing.Event()
        self.pool_stop.clear()

        maxtasksperchild = 10 if 'nt' == os.name else None

        self.pool = multiprocessing.Pool(
            initializer=_worker_initializer,
            maxtasksperchild=maxtasksperchild,
            initargs=(
                self.pool_stop,
                settings.DEBUG,
            )
        )

        # register the shutdown
        settings.FINALIZERS.append(self.shutdown)

    def submit(self, func):
        """
            The callback 'func' will be submitted to a multiprocessing
            pool created by the server. The result object will be returned.
        """
        result = self.pool.apply_async(func)
        self._pyroDaemon.register(result)
        return result
    
    def submit_sync(self, func):
        """
            This will run the 'func' callback directly on the server process.
            Use this with care as it will block the server. 
            Can be used to pass in a full project refinement using the 
            pyxrd.calculations.run_refinement method.
        """
        result = func()
        self._pyroDaemon.register(result)
        return result

    def shutdown(self):
        """
            Shuts down the server.
        """
        # Close the pool:
        logging.info("Closing multiprocessing pool ...")
        if self.pool is not None:
            self.pool_stop.set()
            self.pool.close()
            self.pool.join()
        self.running = False

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

from pyxrd.generic.views import DialogView, BaseView, HasChildView

from pyxrd.goniometer.views import InlineGoniometerView
from pyxrd.generic.views.line_views import CalculatedLinePropertiesView, ExperimentalLinePropertiesView

class SpecimenView(DialogView, HasChildView):
    title = "Edit Specimen"
    subview_builder = resource_filename(__name__, "../glade/specimen.glade")
    subview_toplevel = "edit_specimen"
    resizable = False

    widget_format = "specimen_%s"

    widget_groups = {
        'full_mode_only': [
            "specimen_display_calculated",
            "specimen_display_stats_in_lbl",
            "specimen_display_phases",
            "vbox_calculated_data_tv",
            "lbl_specimen_calculated",
            "vbox_exclusion_ranges_tv",
            "lbl_tabexclusions",
            "general_separator",
            "specimen_display_residuals",
            "specimen_display_derivatives"
        ]
    }

    gonio_container = widget_format % "goniometer"
    gonio_view = None

    calc_line_container = widget_format % "calc_line"
    calc_line_view = None

    exp_line_container = widget_format % "exp_line"
    exp_line_view = None


    def __init__(self, *args, **kwargs):
        super(SpecimenView, self).__init__(*args, **kwargs)
        self._init_child_views()
        self._add_child_views()

    _children_ready = False
    def _init_child_views(self):
        if not self._children_ready:
            self.gonio_view = InlineGoniometerView(parent=self)
            self.calc_line_view = CalculatedLinePropertiesView(parent=self)
            self.exp_line_view = ExperimentalLinePropertiesView(parent=self)
            self._children_ready = True

    def _add_child_views(self):
        self._init_child_views()

        # Add in-line gonio view:
        top = self.gonio_view.get_top_widget()
        self._add_child_view(top, self[self.gonio_container])

        # Add in-line calculated line properties view:
        top = self.calc_line_view.get_top_widget()
        self._add_child_view(top, self[self.calc_line_container])

        # Add in-line experimental line properties view:
        top = self.exp_line_view.get_top_widget()
        self._add_child_view(top, self[self.exp_line_container])

    def set_layout_mode(self, state):
        super(SpecimenView, self).set_layout_mode(state)
        if self._children_ready:
            self.gonio_view.set_layout_mode(state)
            self.calc_line_view.set_layout_mode(state)
            self.exp_line_view.set_layout_mode(state)


    pass # end of class

class StatisticsView(BaseView):
    builder = resource_filename(__name__, "../specimen/glade/statistics.glade")
    top = "statistics_box"

    def __init__(self, *args, **kwargs):
        BaseView.__init__(self, *args, **kwargs)

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

import numpy as np

from pyxrd.calculations.data_objects import GonioData
from pyxrd.calculations.goniometer import (
    get_2t_from_nm,
    get_nm_from_2t, 
    get_nm_from_t,
    get_t_from_nm,
    get_S, 
    get_fixed_to_ads_correction_range,
    get_lorentz_polarisation_factor
)

__all__ = [
    'TestGoniometerCalcs',
]

class TestGoniometerCalcs(unittest.TestCase):

    def setUp(self):
        self.goniometer_data_kwargs = dict(
            min_2theta = 3.0,
            max_2theta = 45,
            mcr_2theta = 0,
            steps = 2500,
            soller1 = 2.3,
            soller2 = 2.3,
            divergence = 0.5,
            has_ads = False,
            ads_fact = 1.0,
            ads_phase_fact = 1.0,
            ads_phase_shift = 0.0,
            ads_const = 0.0,
            radius = 24.0,
            wavelength_distribution = [
                [0.1544426,0.955148885],
                [0.153475,0.044851115],
            ]  
        )
        self.goniometer_data = GonioData(**self.goniometer_data_kwargs)

    def tearDown(self):
        del self.goniometer_data

    def test_not_none(self):
        self.assertIsNotNone(self.goniometer_data)
        
    def test_attributes(self):
        for key, value in self.goniometer_data_kwargs.items():
            self.assertEquals(getattr(self.goniometer_data, key), value)

    def test_fixed_to_ads_correction_range(self):
        result = get_fixed_to_ads_correction_range(
            np.asanyarray([2.2551711385, 2.478038901, 2.7001518288, 2.9214422642, 3.1418428, 3.3612863, 3.5797059197, 3.7970351263]),
            self.goniometer_data
        )
        self.assertIsNotNone(result)
        self.assertEquals(np.allclose(
            result, 
            [7.74814435e-01, 6.15920411e-01, 4.27242611e-01, 2.18376385e-01, -2.50146408e-04, -2.17930643e-01, -4.24231682e-01, -6.09510086e-01]
        ), True)

    def test_lorentz_polarisation_factor(self):
        result = get_lorentz_polarisation_factor(
            np.asanyarray([2.2551711385, 2.478038901, 2.7001518288, 2.9214422642, 3.1418428, 3.3612863, 3.5797059197, 3.7970351263]),
            12,
            self.goniometer_data.soller1, self.goniometer_data.soller2, self.goniometer_data.mcr_2theta
        )
        self.assertIsNotNone(result)
        self.assertEquals(np.allclose(
            result, 
            [3.00643375e-03, 4.83799816e-03, 1.33173586e-02, 6.56714627e-02, 5.11941694e+02, 6.59637604e-02, 1.35695934e-02, 4.97673826e-03]
            #[3.67253512e-01, 5.84371208e-01, 1.55735810e+00, 6.54044238e+00, 7.99531670e+03, 6.56417033e+00, 1.58546668e+00, 6.00742559e-01]
        ), True)
        pass

    def test_S(self):
        # 
        result = get_S(2.5, 2.5);
        self.assertIsNotNone(result)
        self.assertEquals(len(result), 2)
        S, S1S2 = result
        self.assertAlmostEquals(S, 1.7677669529663689)
        self.assertAlmostEquals(S1S2, 6.25)

    def test_2t_from_nm_positive(self):
        # boundary condition: positive
        result = get_2t_from_nm(0.716)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, 12.351779659)
    def test_2t_from_nm_zero(self):
        # boundary condition: 0
        result = get_2t_from_nm(0)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, 0)
    def test_2t_from_nm_negative(self):
        # boundary condition: negative
        result = get_2t_from_nm(-0.716)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, -12.351779659)
     
    def test_nm_from_2t_positive(self):
        # boundary condition: positive
        result = get_nm_from_2t(12.351779659)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, 0.716)
    def test_nm_from_2t_zero(self):
        # boundary condition: 0
        result = get_nm_from_2t(0)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, 1e+16)
        result = get_nm_from_2t(0, zero_for_inf=True)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, 0)
    def test_nm_from_2t_negative(self):
        # boundary condition: negative
        result = get_nm_from_2t(-12.351779659)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, -0.716)
        
    def test_nm_from_t_positive(self):
        # boundary condition: positive
        result = get_nm_from_t(6.17588983)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, 0.716)        
    def test_nm_from_t_zero(self):
        # boundary condition: 0
        result = get_nm_from_t(0)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, 1e+16)
        result = get_nm_from_t(0, zero_for_inf=True)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, 0)
    def test_nm_from_t_negative(self):
        # boundary condition: negative
        result = get_nm_from_t(-6.17588983)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, -0.716)
       
    def test_t_from_nm_positive(self):
        # boundary condition: positive
        result = get_t_from_nm(0.716)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, 6.17588983)
    def test_t_from_nm_zero(self):
        # boundary condition: 0
        result = get_t_from_nm(0)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, 0)
    def test_t_from_nm_negative(self):
        # boundary condition: negative
        result = get_t_from_nm(-0.716)
        self.assertIsNotNone(result)
        self.assertAlmostEquals(result, -6.17588983)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .namespace import xrd_parsers

from .brk_brml_parser import BrkBRMLParser
from .brk_raw_parser import BrkRAWParser
from .cpi_parser import CPIParser
from .csv_parser import CSVParser
from .rd_parser import RDParser
from .udf_parser import UDFParser


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from test.utils import create_object_attribute_test

from pyxrd.atoms.models import AtomType

__all__ = [
    'TestAtomType',
]

class TestAtomType(unittest.TestCase):

    atom_type = None

    def setUp(self):
        self.atom_type = AtomType()

    def tearDown(self):
        del self.atom_type

    def test_not_none(self):
        self.assertIsNotNone(self.atom_type)

    def test_data_object(self):
        self.assertIsNotNone(self.atom_type.data_object)

    test_name = create_object_attribute_test("atom_type", "name", "Test Name")
    test_charge = create_object_attribute_test("atom_type", "charge", -5)
    test_debye = create_object_attribute_test("atom_type", "debye", 1.0)
    test_weight = create_object_attribute_test("atom_type", "weight", 60.123)
    test_atom_nr = create_object_attribute_test("atom_type", "atom_nr", 20)
    test_par_c = create_object_attribute_test("atom_type", "par_c", 10.2)
    test_par_a1 = create_object_attribute_test("atom_type", "par_a1", 10.2)
    test_par_a2 = create_object_attribute_test("atom_type", "par_a2", 10.2)
    test_par_a3 = create_object_attribute_test("atom_type", "par_a3", 10.2)
    test_par_a4 = create_object_attribute_test("atom_type", "par_a4", 10.2)
    test_par_a5 = create_object_attribute_test("atom_type", "par_a5", 10.2)
    test_par_b1 = create_object_attribute_test("atom_type", "par_b1", 10.2)
    test_par_b2 = create_object_attribute_test("atom_type", "par_b2", 10.2)
    test_par_b3 = create_object_attribute_test("atom_type", "par_b3", 10.2)
    test_par_b4 = create_object_attribute_test("atom_type", "par_b4", 10.2)
    test_par_b5 = create_object_attribute_test("atom_type", "par_b5", 10.2)

    def test_parent(self):
        parent_atom_type = AtomType(name="Parent")
        self.atom_type.parent = parent_atom_type
        self.assertEqual(self.atom_type.parent, parent_atom_type)
        
    def test_get_atomic_scattering_factors(self):
        import numpy as np
        stl_range = np.array([ 0.1152, 0.5756, 1.1469 ])
        self.atom_type.par_a1 = 10.2
        self.atom_type.par_a2 = 10.2
        self.atom_type.par_a3 = 10.2
        self.atom_type.par_a4 = 10.2
        self.atom_type.par_a5 = 10.2
        self.atom_type.par_b1 = 10.2
        self.atom_type.par_b2 = 10.2
        self.atom_type.par_b3 = 10.2
        self.atom_type.par_b4 = 10.2
        self.atom_type.par_b5 = 10.2
        self.atom_type.par_c = 20.2
        self.assertListEqual(
            self.atom_type.get_atomic_scattering_factors(stl_range).tolist(),
            [ 71.1827439324707, 70.77093939463964, 69.51772020477823 ]
        )

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2016 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import unittest

from mvc.adapters.gtk_support.dialogs.file_chooser_dialog import FileChooserDialog

from .file_chooser_dialog_args import get_file_chooser_kwags
from mock.mock import Mock

__all__ = [
    'FileChooserDialogTest',
]

class FileChooserDialogTest(unittest.TestCase):

    def setUp(self):
        self.kwargs = get_file_chooser_kwags()
        self.dialog = FileChooserDialog(**self.kwargs)


    def tearDown(self):
        pass


    def test_file_dialog_selected_globs(self):
        
        # Set filter:
        filter = self.dialog.list_filters()[0] #this is a *.txt filter, at least it should be... @ReservedAssignment
        self.dialog.set_filter(filter)
        
        #Act        
        sel_globs = self.dialog.selected_globs
        
        #Assert
        self.assertEqual(filter.get_name(), "Text File")
        self.assertEqual(sel_globs, ["*.txt"])
        
        

    def test_file_dialog_applies_filename_filters(self):
        
        no_ext_name = "test name without extension"
        
        # Set filter:
        filter = self.dialog.list_filters()[0] #this is a *.txt filter, at least it should be... @ReservedAssignment
        self.dialog.set_filter(filter)
        # Mock current filename
        self.dialog.get_filename = Mock(return_value=no_ext_name)
        
        # Assert we are working with the filter we expect
        self.assertEqual(filter.get_name(), "Text File")
        
        # Assert name is changed correctly
        self.assertEqual("%s.txt" % no_ext_name, self.dialog.filename)
        self.dialog.get_filename.assert_called_once_with()

if __name__ == "__main__":
    unittest.main()


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import numpy as np

from mvc.models.xydata import XYData

from pyxrd.generic.io import storables, Storable

from pyxrd.generic.models.base import DataModel
from pyxrd.generic.utils import not_none

#from pyxrd.file_parsers.ascii_parser import ASCIIParser

@storables.register()
class StorableXYData(DataModel, XYData, Storable):
    """
        A storable XYData model with additional I/O and CRUD abilities.
    """

    class Meta(XYData.Meta):
        store_id = "StorableXYData"

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Valid keyword arguments for a PyXRDLine are:
                data: the actual data containing x and y values
                label: the label for this line
                color: the color of this line
                inherit_color: whether to use the parent-level color or its own
                lw: the line width of this line
                inherit_lw: whether to use the parent-level line width or its own
        """
        if "xy_store" in kwargs:
            kwargs["data"] = kwargs.pop("xy_store")
        super(StorableXYData, self).__init__(*args, **kwargs)

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def json_properties(self):
        props = super(XYData, self).json_properties()
        props["data"] = self._serialize_data()
        return props

    def apply_correction(self, correction):
        self.data_y = self.data_y * correction[:, np.newaxis]

    def save_data(self, parser, filename, **kwargs):
        if self.data_y.shape[1] > 1:
            kwargs["header"] = ["2θ", ] + (not_none(self.y_names, []))
        parser.write(filename, self.data_x, self._data_y.transpose(), **kwargs)

    def load_data(self, parser, filename, clear=True):
        """
            Loads data using passed filename and parser, which are passed on to
            the load_data_from_generator method.
            If clear=True the x-y data is cleared first.
        """
        xrdfiles = parser.parse(filename)
        if xrdfiles:
            self.load_data_from_generator(xrdfiles[0].data, clear=clear)
            
    def load_data_from_generator(self, generator, clear=True):
        with self.data_changed.hold_and_emit():
            with self.visuals_changed.hold_and_emit():
                super(StorableXYData, self).load_data_from_generator(generator, clear=clear)

    def set_data(self, x, y):
        """
            Sets data using the supplied x, y1, ..., yn arrays.
        """
        with self.data_changed.hold_and_emit():
            with self.visuals_changed.hold_and_emit():
                super(StorableXYData, self).set_data(x, y)

    def set_value(self, i, j, value):
        with self.data_changed.hold_and_emit():
            with self.visuals_changed.hold_and_emit():
                super(StorableXYData, self).set_value(i, j, value)

    def append(self, x, y):
        """
            Appends data using the supplied x, y1, ..., yn arrays.
        """
        with self.data_changed.hold_and_emit():
            with self.visuals_changed.hold_and_emit():
                super(StorableXYData, self).append(x, y)

    def insert(self, pos, x, y):
        """
            Inserts data using the supplied x, y1, ..., yn arrays at the given
            position.
        """
        with self.data_changed.hold_and_emit():
            with self.visuals_changed.hold_and_emit():
                super(StorableXYData, self).insert(pos, x, y)

    def remove_from_indeces(self, *indeces):
        with self.data_changed.hold_and_emit():
            with self.visuals_changed.hold_and_emit():
                super(StorableXYData, self).remove_from_indeces(*indeces)

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

toolkit = "gtk"

def load(tkreg):
    # Load adapters:
    from .adjustment_adapter import AdjustmentAdapter
    from .arrow_adapter import ArrowAdapter
    from .check_menu_item_adapter import CheckMenuItemAdapter
    from .color_button_adapter import ColorButtonAdapter
    from .color_selection_adapter import ColorSelectionAdapter
    from .combo_box_adapter import ComboBoxAdapter
    from .entry_adapter import EntryAdapter
    from .expander_adapter import ExpanderAdapter
    from .file_chooser_adapter import FileChooserAdapter
    from .float_entry_adapter import FloatEntryAdapter
    from .label_adapter import LabelAdapter
    from .link_button_adapter import LinkButtonAdapter
    from .scale_adapter import ScaleEntryAdapter
    from .text_view_adapter import TextViewAdapter
    from .toggle_button_adapter import ToggleButtonAdapter
    from .tree_view_adapters import XYListViewAdapter, ObjectListViewAdapter
    
    # Load main loop functions
    from .toolkit_functions import (
        add_idle_call, remove_source, 
        add_timeout_call, 
        start_event_loop, stop_event_loop
    )
    
    # Register them
    tkreg.set_toolkit_functions(
        add_idle_call, remove_source, 
        add_timeout_call, remove_source, 
        start_event_loop, stop_event_loop
    )



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import inspect
import logging
from mvc.support.gui_loop import add_idle_call

logger = logging.getLogger(__name__)

from weakref import WeakKeyDictionary

try:
    import threading as threading
except ImportError: 
    import dummy_threading as threading
    
try:
    from fastrlock.rlock import FastRLock as RLock
except ImportError:    
    from threading import RLock

from ..support.collections.weak_list import WeakList

from ..support.observables import ObsWrapperBase, Signal
from ..observers import Observer, NTInfo

from .metaclasses import ModelMeta
from .properties import UUIDProperty

class Model(Observer, metaclass=ModelMeta):
    """
    .. attribute:: __observables__
    
       Class attribute. A list or tuple of name strings. The metaclass
       :class:`~mvc.support.metaclasses.ObservablePropertyMeta`
       uses it to create properties.
       
       *Value properties* have to exist as an attribute with an
       initial value, which may be ``None``.

       *Logical properties* require a getter and may have a setter method in
       the class.
    """

    """A base class for models whose observable properties can be
    changed by threads different than the (gtk) main thread. Notification is
    performed by exploiting the gtk idle loop only if needed,
    otherwise the standard notification system (direct method call) is
    used. In this model, the observer is expected to run in the gtk
    main loop thread."""

    class Meta(object):
        """
            A meta-data class providing some basic functionality 
        """

        @classmethod
        def get_column_properties(cls):
            if not hasattr(cls, "all_properties"):
                raise RuntimeError("Meta class '%s' has not been initialized" \
                    " properly: 'all_properties' is not set!" % type(cls))
            else:
                cls._mem_columns = getattr(cls, "_mem_columns", None)
                if cls._mem_columns is None:
                    cls._mem_columns = [(attr.label, attr.data_type) for attr in cls.all_properties if attr.tabular]
                return cls._mem_columns

        @classmethod
        def get_local_persistent_properties(cls):
            return [attr for attr in cls.properties if attr.persistent]

        @classmethod
        def get_viewless_properties(cls):
            if not hasattr(cls, "all_properties"):
                raise RuntimeError("Meta class '%s' has not been initialized" \
                    " properly: 'all_properties' is not set!" % type(self))
            else:
                return [attr for attr in cls.all_properties if not attr.visible]

        @classmethod
        def get_viewable_properties(cls):
            if not hasattr(cls, "all_properties"):
                raise RuntimeError("Meta class '%s' has not been initialized" \
                    " properly: 'all_properties' is not set!" % type(self))
            else:
                return [attr for attr in cls.all_properties if attr.visible]

        pass # end of class

    uuid = UUIDProperty(persistent=True, observable=False)

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)

        self._prop_lock = RLock() # @UndefinedVariable
        self.__observers = WeakList()
        self.__observer_threads = WeakKeyDictionary()

        # keys are properties names, values are pairs (method,
        # kwargs|None) inside the observer. kwargs is the keyword
        # argument possibly specified when explicitly defining the
        # notification method in observers, and it is used to build
        # the NTInfo instance passed down when the notification method
        # is invoked. If kwargs is None (special case), the
        # notification method is "old style" (property_<name>_...) and
        # won't be receiving the property name.
        self.__value_notifications = {}
        self.__instance_notif_before = {}
        self.__instance_notif_after = {}
        self.__signal_notif = {}

        for attr in self.Meta.all_properties: self.register_property(attr)
        return

    def register_property(self, prop):
        """Registers an existing attribute to be monitored, and sets
        up notifiers for notifications"""
        if prop.label not in self.__value_notifications:
            self.__value_notifications[prop.label] = []
            pass

        # registers observable wrappers
        propval = getattr(type(self), prop.label)._get(self)

        if isinstance(propval, ObsWrapperBase):
            propval.__add_model__(self, prop.label)

            if isinstance(propval, Signal):
                if prop.label not in self.__signal_notif:
                    self.__signal_notif[prop.label] = []
                    pass
                pass
            else:
                if prop.label not in self.__instance_notif_before:
                    self.__instance_notif_before[prop.label] = []
                    pass
                if prop.label not in self.__instance_notif_after:
                    self.__instance_notif_after[prop.label] = []
                    pass
                pass
            pass

        return

    def has_property(self, label):
        """Returns true if given property name refers an observable
        property inside self or inside derived classes."""
        for prop in self.Meta.all_properties:
            if prop.label == label:
                return True

    def register_observer(self, observer):
        """Register given observer among those observers which are
        interested in observing the model."""
        if observer in self.__observers: return # not already registered

        assert isinstance(observer, Observer)
        self.__observers.append(observer)
        self.__observer_threads[observer] = threading.current_thread() # @UndefinedVariable
        for prop in self.Meta.all_properties:
            self.__add_observer_notification(observer, prop)
            pass

        return

    def unregister_observer(self, observer):
        """Unregister the given observer that is no longer interested
        in observing the model."""

        assert isinstance(observer, Observer)

        if observer not in self.__observers: return
        for prop in self.Meta.all_properties:
            self.__remove_observer_notification(observer, prop)
            pass

        self.__observers.remove(observer)
        del self.__observer_threads[observer]

        return

    def _reset_property_notification(self, prop, old=None):
        """Called when an assignment has been done that changes the
        type of a property or the instance of the property has been
        changed to a different instance. In this case it must be
        unregistered and registered again. Optional parameter old has
        to be used when the old value is an instance (derived from 
        ObsWrapperBase) which needs to unregistered from the model, via
        a call to method old.__remove_model__(model, prop_name)"""

        # unregister_property
        if isinstance(old, ObsWrapperBase):
            old.__remove_model__(self, prop.label)
            pass

        self.register_property(prop)

        for observer in self.__observers:
            self.__remove_observer_notification(observer, prop)
            self.__add_observer_notification(observer, prop)
            pass
        return

    def __add_observer_notification(self, observer, prop):
        """
        Find observing methods and store them for later notification.

        *observer* an instance.
        
        *label* a string.

        This checks for magic names as well as methods explicitly added through
        decorators or at runtime. In the latter case the type of the notification
        is inferred from the number of arguments it takes.
        """
        value = getattr(type(self), prop.label)._get(self)

        # --- Some services ---
        def getmeth(format, numargs): # @ReservedAssignment
            name = format % prop.label
            meth = getattr(observer, name)
            args, varargs, _, _ = inspect.getargspec(meth)
            if not varargs and len(args) != numargs:
                logger.warn("Ignoring notification %s: exactly %d arguments"
                    " are expected", name, numargs)
                raise AttributeError
            return meth

        def add_value(notification, kw=None):
            pair = (notification, kw)
            if pair in self.__value_notifications[prop.label]: return
            logger.debug("Will call %s.%s after assignment to %s.%s",
                observer.__class__.__name__, notification.__name__,
                self.__class__.__name__, prop.label)
            self.__value_notifications[prop.label].append(pair)
            return

        def add_before(notification, kw=None):
            if (not isinstance(value, ObsWrapperBase) or
                isinstance(value, Signal)):
                return

            pair = (notification, kw)
            if pair in self.__instance_notif_before[prop.label]: return
            logger.debug("Will call %s.%s before mutation of %s.%s",
                observer.__class__.__name__, notification.__name__,
                self.__class__.__name__, prop.label)
            self.__instance_notif_before[prop.label].append(pair)
            return

        def add_after(notification, kw=None):
            if (not isinstance(value, ObsWrapperBase) or
                isinstance(value, Signal)):
                return
            pair = (notification, kw)
            if pair in self.__instance_notif_after[prop.label]: return
            logger.debug("Will call %s.%s after mutation of %s.%s",
                observer.__class__.__name__, notification.__name__,
                self.__class__.__name__, prop.label)
            self.__instance_notif_after[prop.label].append(pair)
            return

        def add_signal(notification, kw=None):
            if not isinstance(value, Signal): return
            pair = (notification, kw)
            if pair in self.__signal_notif[prop.label]: return
            logger.debug("Will call %s.%s after emit on %s.%s",
                observer.__class__.__name__, notification.__name__,
                self.__class__.__name__, prop.label)
            self.__signal_notif[prop.label].append(pair)
            return
        # ---------------------

        try: notification = getmeth("property_%s_signal_emit", 3)
        except AttributeError: pass
        else: add_signal(notification)

        try: notification = getmeth("property_%s_value_change", 4)
        except AttributeError: pass
        else: add_value(notification)

        try: notification = getmeth("property_%s_before_change", 6)
        except AttributeError: pass
        else: add_before(notification)

        try: notification = getmeth("property_%s_after_change", 7)
        except AttributeError: pass
        else: add_after(notification)

        # here explicit notification methods are handled (those which
        # have been statically or dynamically registered)
        type_to_adding_method = {
            'assign' : add_value,
            'before' : add_before,
            'after'  : add_after,
            'signal' : add_signal,
            }

        for meth in observer.get_observing_methods(prop.label):
            added = False
            kw = observer.get_observing_method_kwargs(prop.label, meth)
            for flag, adding_meth in type_to_adding_method.items():
                if flag in kw:
                    added = True
                    adding_meth(meth, kw)
                    pass
                pass
            if not added: raise ValueError("In %s notification method %s is "
                                           "marked to be observing property "
                                           "'%s', but no notification type "
                                           "information were specified." %
                                           (observer.__class__,
                                            meth.__name__, prop.label))
            pass

        return

    def __remove_observer_notification(self, observer, prop):
        """
        Remove all stored notifications.
        
        *observer* an instance.
        
        *prop* a LabeledProperty instance.
        """

        def side_effect(seq):
            for meth, kw in reversed(seq):
                if meth.__self__ is observer:
                    seq.remove((meth, kw))
                    yield meth

        for meth in side_effect(self.__value_notifications.get(prop.label, ())):
            logger.debug("Stop calling '%s' after assignment", meth.__name__)

        for meth in side_effect(self.__signal_notif.get(prop.label, ())):
            logger.debug("Stop calling '%s' after emit", meth.__name__)

        for meth in side_effect(self.__instance_notif_before.get(prop.label, ())):
            logger.debug("Stop calling '%s' before mutation", meth.__name__)

        for meth in side_effect(self.__instance_notif_after.get(prop.label, ())):
            logger.debug("Stop calling '%s' after mutation", meth.__name__)

        return

    def __notify_observer__(self, observer, method, *args, **kwargs):
        """This makes a call either through the Gtk.idle list or a
        direct method call depending whether the caller's thread is
        different from the observer's thread"""

        assert observer in self.__observer_threads
        if threading.currentThread() == self.__observer_threads[observer]: # @UndefinedVariable
            self.__idle_notify_observer(observer, method, args, kwargs)
        else:
            add_idle_call(self.__idle_notify_observer, observer, method, args, kwargs)

    def __idle_notify_observer(self, observer, method, args, kwargs):
        method(*args, **kwargs)


    # -------------------------------------------------------------
    #            Notifiers:
    # -------------------------------------------------------------
    def notify_property_value_change(self, prop_name, old, new):
        """
        Send a notification to all registered observers.

        *old* the value before the change occured.
        """
        assert(prop_name in self.__value_notifications)
        for method, kw in self.__value_notifications[prop_name] :
            obs = method.__self__
            # notification occurs checking spuriousness of the observer
            if old != new or obs.accepts_spurious_change():
                if kw is None: # old style call without name
                    self.__notify_observer__(obs, method,
                                             self, old, new)
                elif 'old_style_call' in kw:  # old style call with name
                    self.__notify_observer__(obs, method,
                                             self, prop_name, old, new)
                else:
                    # New style explicit notification.
                    # notice that named arguments overwrite any
                    # existing key:val in kw, which is precisely what
                    # it is expected to happen
                    info = NTInfo('assign',
                                  kw, model=self, prop_name=prop_name,
                                  old=old, new=new)
                    self.__notify_observer__(obs, method,
                                             self, prop_name, info)
                    pass
                pass
            pass
        return

    def notify_method_before_change(self, prop_name, instance, meth_name,
                                    args, kwargs):
        """
        Send a notification to all registered observers.

        *instance* the object stored in the property.

        *meth_name* name of the method we are about to call on *instance*.
        """
        assert(prop_name in self.__instance_notif_before)
        for method, kw in self.__instance_notif_before[prop_name]:
            obs = method.__self__
            # notifies the change
            if kw is None: # old style call without name
                self.__notify_observer__(obs, method,
                                         self, instance,
                                         meth_name, args, kwargs)
            elif 'old_style_call' in kw:  # old style call with name
                self.__notify_observer__(obs, method,
                                         self, prop_name, instance,
                                         meth_name, args, kwargs)
            else:
                # New style explicit notification.
                # notice that named arguments overwrite any
                # existing key:val in kw, which is precisely what
                # it is expected to happen
                info = NTInfo('before',
                              kw,
                              model=self, prop_name=prop_name,
                              instance=instance, method_name=meth_name,
                              args=args, kwargs=kwargs)
                self.__notify_observer__(obs, method,
                                         self, prop_name, info)
                pass
            pass
        return

    def notify_method_after_change(self, prop_name, instance, meth_name,
                                   res, args, kwargs):
        """
        Send a notification to all registered observers.

        *args* the arguments we just passed to *meth_name*.

        *res* the return value of the method call.
        """
        assert(prop_name in self.__instance_notif_after)
        for method, kw in self.__instance_notif_after[prop_name]:
            obs = method.__self__
            # notifies the change
            if kw is None:  # old style call without name
                self.__notify_observer__(obs, method,
                                         self, instance,
                                         meth_name, res, args, kwargs)
            elif 'old_style_call' in kw:  # old style call with name
                self.__notify_observer__(obs, method,
                                         self, prop_name, instance,
                                         meth_name, res, args, kwargs)
            else:
                # New style explicit notification.
                # notice that named arguments overwrite any
                # existing key:val in kw, which is precisely what
                # it is expected to happen
                info = NTInfo('after',
                              kw,
                              model=self, prop_name=prop_name,
                              instance=instance, method_name=meth_name,
                              result=res, args=args, kwargs=kwargs)
                self.__notify_observer__(obs, method,
                                         self, prop_name, info)
                pass
            pass
        return

    def notify_signal_emit(self, prop_name, arg):
        """
        Emit a signal to all registered observers.

        *prop_name* the property storing the :class:`~mvc.observable.Signal`
        instance.

        *arg* one arbitrary argument passed to observing methods.
        """
        assert(prop_name in self.__signal_notif)

        for method, kw in self.__signal_notif[prop_name]:
            obs = method.__self__
            # notifies the signal emit
            if kw is None: # old style call, without name
                self.__notify_observer__(obs, method,
                                         self, arg)
            elif 'old_style_call' in kw:  # old style call with name
                self.__notify_observer__(obs, method,
                                         self, prop_name, arg)
            else:
                # New style explicit notification.
                # notice that named arguments overwrite any
                # existing key:val in kw, which is precisely what
                # it is expected to happen
                info = NTInfo('signal',
                              kw,
                              model=self, prop_name=prop_name, arg=arg)
                self.__notify_observer__(obs, method,
                                         self, prop_name, info)
                pass
            pass
        return


    pass # end of class Model
# ----------------------------------------------------------------------



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .base import AbstractTestProbModel

from test.utils import create_object_attribute_test

from pyxrd.probabilities.models.R2models import *
import unittest

__all__ = [
    'TestR2G2Model',
    'TestR2G3Model',
]

class TestR2G2Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R2G2Model

    test_W1 = create_object_attribute_test("prob_model", "W1", 0.7)
    test_P112_or_P211 = create_object_attribute_test("prob_model", "P112_or_P211", 0.7)
    test_P21 = create_object_attribute_test("prob_model", "P21", 0.7)
    test_P122_or_P221 = create_object_attribute_test("prob_model", "P122_or_P221", 0.7)

    pass # end of class

class TestR2G3Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R2G3Model

    test_W1 = create_object_attribute_test("prob_model", "W1", 0.7)
    test_P111_or_P212 = create_object_attribute_test("prob_model", "P111_or_P212", 0.7)
    test_G1 = create_object_attribute_test("prob_model", "G1", 0.7)
    test_G2 = create_object_attribute_test("prob_model", "G2", 0.7)
    test_G3 = create_object_attribute_test("prob_model", "G3", 0.7)
    test_G4 = create_object_attribute_test("prob_model", "G4", 0.7)

    pass # end of class
